#!/usr/bin/python3

r"""Class to read and manage relevant model parameters for a given telescope model."""

import logging
import re
import textwrap
from collections import defaultdict
from itertools import groupby
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from simtools.db import db_handler
from simtools.io import ascii_handler, io_handler
from simtools.model.telescope_model import TelescopeModel
from simtools.utils import names
from simtools.version import sort_versions
from simtools.visualization import plot_pixels, plot_tables

logger = logging.getLogger()
MARKDOWN_LINK_RE = re.compile(r"\[([^\]]+)\]\(([^)]+)\)")


class ReadParameters:
    """Read and manage model parameter data for report generation."""

    def __init__(self, db_config, args, output_path):
        """Initialise class."""
        self._logger = logging.getLogger(__name__)
        self.db = db_handler.DatabaseHandler(db_config=db_config)
        self.db_config = db_config
        self.array_element = args.get("telescope", None)
        self.site = args.get("site", None)
        self.model_version = args.get("model_version", None)
        self.output_path = output_path
        self.observatory = args.get("observatory")
        self.software = args.get("simulation_software", None)

    @property
    def model_version(self):
        """Model version."""
        return self._model_version

    @model_version.setter
    def model_version(self, model_version):
        """
        Set model version.

        Parameters
        ----------
        _model_version: str or list
            Model version (e.g., "6.0.0").
            If a list is passed, it must contain exactly one element,
            and only that element will be used.

        Raises
        ------
        ValueError
            If more than one model version is passed.
        """
        if isinstance(model_version, list):
            raise ValueError(
                f"Only one model version can be passed to {self.__class__.__name__}, not a list."
            )
        self._model_version = model_version

    def _generate_plots(self, parameter, parameter_version, input_file, outpath):
        """Generate plots based on the parameter type."""
        plot_names = []

        if parameter == "camera_config_file":
            plot_names = self._plot_camera_config(parameter, parameter_version, input_file, outpath)
        elif parameter_version:
            plot_names = self._plot_parameter_tables(
                parameter,
                parameter_version,
                outpath,
            )

        return plot_names

    def _plot_camera_config(self, parameter, parameter_version, input_file, outpath):
        """Generate plots for camera configuration files."""
        if not parameter_version:
            return []

        plot_names = []
        plot_name = input_file.stem.replace(".", "-")
        plot_path = Path(f"{outpath}/{plot_name}").with_suffix(".png")

        if not plot_path.exists():
            plot_config = {
                "file_name": input_file.name,
                "telescope": self.array_element,
                "parameter_version": parameter_version,
                "site": self.site,
                "model_version": self.model_version,
                "parameter": parameter,
            }

            plot_pixels.plot(
                config=plot_config,
                output_file=Path(f"{outpath}/{plot_name}"),
                db_config=self.db_config,
            )
            plot_names.append(plot_name)
        else:
            logger.info(
                "Camera configuration file plot already exists: %s",
                plot_path,
            )
            plot_names.append(plot_name)

        return plot_names

    def _plot_parameter_tables(self, parameter, parameter_version, outpath):
        """Generate plots for parameter tables."""
        tel = self._get_telescope_identifier()

        config_data = plot_tables.generate_plot_configurations(
            parameter=parameter,
            parameter_version=parameter_version,
            site=self.site,
            telescope=tel,
            output_path=outpath,
            plot_type="all",
            db_config=self.db_config,
        )

        if not config_data:
            return []

        plot_configs, output_files = config_data
        plot_names = [i.stem for i in output_files]

        for plot_config, output_file in zip(plot_configs, output_files):
            image_output_file = outpath / output_file.name
            if not image_output_file.with_suffix(".png").exists():
                plot_tables.plot(
                    config=plot_config,
                    output_file=image_output_file,
                    db_config=self.db_config,
                )
                plt.close("all")

        return plot_names

    def _get_telescope_identifier(self, model_version=None):
        """Get the appropriate telescope design type for file naming (e.g., LSTN-design)."""
        model_version = model_version or self.model_version
        telescope_design = self.db.get_design_model(
            model_version, self.array_element, collection="telescopes"
        )

        if not self.array_element:
            return None
        if not names.is_design_type(self.array_element):
            return telescope_design
        return self.array_element

    def _convert_to_md(self, parameter, parameter_version, input_file):
        """Convert a file to a Markdown file, preserving formatting."""
        input_file = Path(input_file)

        if not input_file.exists():
            raise FileNotFoundError(f"Data file not found: {input_file}")

        # Store the markdown output file path early and don't modify it
        output_data_path = Path(self.output_path / "_data_files")
        output_data_path.mkdir(parents=True, exist_ok=True)
        output_file_name = Path(input_file.stem + ".md")
        relative_path = f"_data_files/{output_file_name}"
        markdown_output_file = output_data_path / output_file_name

        if not markdown_output_file.exists():
            outpath = Path(io_handler.IOHandler().get_output_directory().parent / "_images")
            outpath.mkdir(parents=True, exist_ok=True)

            plot_names = self._generate_plots(parameter, parameter_version, input_file, outpath)
            # Write markdown file using the stored path
            file_contents = ascii_handler.read_file_encoded_in_utf_or_latin(input_file)

            with markdown_output_file.open("w", encoding="utf-8") as outfile:
                outfile.write(f"# {input_file.stem}\n")

                for plot_name in plot_names:
                    outfile.write(f"![Parameter plot.]({outpath}/{plot_name}.png)\n\n")

                outfile.write(
                    "\n\nThe full file can be found in the Simulation Model repository [here]"
                    "(https://gitlab.cta-observatory.org/cta-science/simulations/"
                    "simulation-model/simulation-models/-/blob/main/simulation-models/"
                    f"model_parameters/Files/{input_file.name}).\n\n"
                )
                outfile.write("\n\n")
                outfile.write("The first 30 lines of the file are:\n")
                outfile.write("```\n")
                first_30_lines = "".join(file_contents[:30])
                outfile.write(first_30_lines)
                outfile.write("\n```")

        return relative_path

    def _format_parameter_value(
        self, parameter, value_data, unit, file_flag, parameter_version=None
    ):
        """Format parameter value based on type."""
        if file_flag:
            input_file_name = f"{self.output_path}/model/{value_data}"
            if parameter_version is None:
                return (
                    f"[{Path(value_data).name}](https://gitlab.cta-observatory.org/"
                    "cta-science/simulations/simulation-model/simulation-models/-/blob/main/"
                    f"simulation-models/model_parameters/Files/{value_data})"
                ).strip()
            output_file_name = self._convert_to_md(parameter, parameter_version, input_file_name)
            return f"[{Path(value_data).name}]({output_file_name})".strip()
        if isinstance(value_data, (str | int | float)):
            return f"{value_data} {unit}".strip()
        if len(value_data) > 5 and np.allclose(value_data, value_data[0]):
            return f"all: {value_data[0]} {unit}".strip()
        return (
            ", ".join(f"{v} {u}" for v, u in zip(value_data, unit))
            if isinstance(unit, list)
            else ", ".join(f"{v} {unit}" for v in value_data)
        ).strip()

    def _group_model_versions_by_parameter_version(self, grouped_data):
        """Group model versions by parameter version and track the parameter values."""
        result = {}

        for parameter_name, items in grouped_data.items():
            version_grouped = defaultdict(
                lambda: {"model_versions": [], "value": None, "file_flag": None}
            )

            for item in items:
                param_version = item["parameter_version"]
                version_grouped[param_version]["model_versions"].append(item["model_version"])

                if version_grouped[param_version]["value"] is None:
                    version_grouped[param_version]["value"] = item["value"]
                    version_grouped[param_version]["file_flag"] = item["file_flag"]

            result[parameter_name] = [
                {
                    "value": data["value"],
                    "parameter_version": param_version,
                    "file_flag": data["file_flag"],
                    "model_version": ", ".join(sort_versions(data["model_versions"])),
                }
                for param_version, data in version_grouped.items()
            ]

        return result

    def _compare_parameter_across_versions(self, all_param_data, all_parameter_names):
        """
        Compare a parameter's value across different model versions.

        Parameters
        ----------
        all_param_data : dict
            The dictionary containing parameter data for all versions.

        all_parameter_names : list
            The list of parameter names to compare across versions.

        Returns
        -------
        list
            A list of dictionaries containing model version, parameter value, description.
        """
        all_versions = self.db.get_model_versions()
        all_versions.reverse()  # latest first
        grouped_data = defaultdict(list)

        # Iterate over each model version
        for version in all_versions:
            parameter_dict = all_param_data.get(version, {})
            if not parameter_dict:
                continue

            Path(f"{self.output_path}/model").mkdir(parents=True, exist_ok=True)
            self.db.export_model_files(
                parameters=all_param_data.get(version), dest=f"{self.output_path}/model"
            )

            for parameter_name in filter(parameter_dict.__contains__, all_parameter_names):
                parameter_data = parameter_dict.get(parameter_name)

                # Skip if instrument doesn't match
                if parameter_data.get("instrument") != self.array_element:
                    continue

                unit = parameter_data.get("unit") or " "
                value_data = parameter_data.get("value")

                if value_data is None:
                    continue

                file_flag = parameter_data.get("file", False)
                parameter_version = parameter_data.get("parameter_version")
                value = self._format_parameter_value(
                    parameter_name, value_data, unit, file_flag, parameter_version=None
                )
                model_version = version

                # Group the data by parameter version and store model versions as a list
                grouped_data[parameter_name].append(
                    {
                        "value": value,
                        "parameter_version": parameter_version,
                        "model_version": model_version,
                        "file_flag": file_flag,
                    }
                )

        return self._group_model_versions_by_parameter_version(grouped_data)

    def get_all_parameter_descriptions(self, collection="telescopes"):
        """Get descriptions for all model parameters.

        Returns
        -------
        dict
            Nested dictionaries with first key as the parameter name and
            the following dictionary as the value:
            - key: description, value: description of the parameter.
            - key: short_description, value: short description of the parameter.
            - key: inst_class, value: class, for eg. Structure, Camera, etc.
        """
        parameter_dict = {}

        for instrument_class in names.db_collection_to_instrument_class_key(collection):
            for parameter, details in names.model_parameters(instrument_class).items():
                parameter_dict[parameter] = {
                    "description": details.get("description"),
                    "short_description": details.get("short_description"),
                    "inst_class": instrument_class,
                }

        return parameter_dict

    def get_array_element_parameter_data(self, telescope_model, collection="telescopes"):
        """
        Get model parameter data and descriptions for a given array element.

        Currently only configured for telescope.

        Parameters
        ----------
        telescope_model : TelescopeModel
            The telescope model instance.

        Returns
        -------
        list: A list of lists containing parameter names, values with units,
                descriptions, and short descriptions.
        """
        all_parameter_data = self.db.get_model_parameters(
            site=telescope_model.site,
            array_element_name=telescope_model.name,
            collection=collection,
            model_version=telescope_model.model_version,
        )

        Path(f"{self.output_path}/model").mkdir(parents=True, exist_ok=True)
        self.db.export_model_files(parameters=all_parameter_data, dest=f"{self.output_path}/model")
        parameter_descriptions = self.get_all_parameter_descriptions()
        data = []

        for parameter in filter(all_parameter_data.__contains__, names.model_parameters().keys()):
            parameter_data = all_parameter_data.get(parameter)
            parameter_version = telescope_model.get_parameter_version(parameter)
            unit = parameter_data.get("unit") or " "
            value_data = parameter_data.get("value")

            if value_data is None:
                continue

            file_flag = parameter_data.get("file", False)
            value = self._format_parameter_value(
                parameter, value_data, unit, file_flag, parameter_version
            )

            description = parameter_descriptions.get(parameter).get("description")
            short_description = (
                parameter_descriptions.get(parameter).get("short_description") or description
            )
            inst_class = parameter_descriptions.get(parameter).get("inst_class")

            matching_instrument = parameter_data["instrument"] == telescope_model.name
            if not names.is_design_type(telescope_model.name) and matching_instrument:
                parameter = f"***{parameter}***"
                parameter_version = f"***{parameter_version}***"
                if not self.is_markdown_link(value):
                    value = f"***{value}***"
                description = f"***{description}***"
                short_description = f"***{short_description}***"

            data.append(
                [
                    inst_class,
                    parameter,
                    parameter_version,
                    value,
                    description,
                    short_description,
                ]
            )

        return data

    def _write_to_file(self, data, file):
        # Write table header and separator row
        file.write(
            "| Parameter Name      |  Parameter Version     "
            "| Values      | Short Description           |\n"
            "|---------------------|------------------------"
            "|-------------|-----------------------------|\n"
        )

        # Write table rows
        column_widths = [10, 10, 20, 60]
        for (
            _,
            parameter_name,
            parameter_version,
            value,
            description,
            short_description,
        ) in data:
            text = short_description if short_description else description
            wrapped_text = textwrap.fill(str(text), column_widths[3]).split("\n")
            wrapped_text = " ".join(wrapped_text)

            file.write(
                f"| {parameter_name:{column_widths[0]}} |"
                f" {parameter_version:{column_widths[1]}} |"
                f" {value:{column_widths[2]}} |"
                f" {wrapped_text:{column_widths[3]}} |\n"
            )
        file.write("\n\n")

    def get_simulation_configuration_data(self):
        """Get data and descriptions for simulation configuration parameters."""

        def get_param_data(telescope, site):
            """Retrieve and format parameter data for one telescope-site combo."""
            param_dict = self.db.get_simulation_configuration_parameters(
                simulation_software=self.software,
                site=site,
                array_element_name=telescope,
                model_version=self.model_version,
            )

            parameter_descriptions = self.get_all_parameter_descriptions(
                collection=f"configuration_{self.software}"
            )

            model_output_path = Path(f"{self.output_path}/model")
            model_output_path.mkdir(parents=True, exist_ok=True)
            self.db.export_model_files(parameters=param_dict, dest=str(model_output_path))

            data = []
            for parameter, parameter_data in param_dict.items():
                description = parameter_descriptions.get(parameter).get("description")
                short_description = parameter_descriptions.get(parameter).get(
                    "short_description", description
                )
                value_data = parameter_data.get("value")

                if value_data is None:
                    continue

                unit = parameter_data.get("unit") or " "
                file_flag = parameter_data.get("file", False)
                parameter_version = parameter_data.get("parameter_version")
                value = self._format_parameter_value(
                    parameter, value_data, unit, file_flag, parameter_version
                )

                data.append(
                    [
                        telescope,
                        parameter,
                        parameter_version,
                        value,
                        description,
                        short_description,
                    ]
                )
            return data

        if self.software == "corsika":
            return get_param_data(self.array_element, self.site)

        results = []
        telescopes = self.db.get_array_elements(self.model_version)
        for telescope in telescopes:
            valid_site = names.get_site_from_array_element_name(telescope)
            if not isinstance(valid_site, list):
                results.extend(get_param_data(telescope, valid_site))
            else:
                for site in valid_site:
                    results.extend(get_param_data(telescope, site))
        return results

    def produce_simulation_configuration_report(self):
        """Write simulation configuration report."""
        output_filename = Path(self.output_path / (f"configuration_{self.software}.md"))
        output_filename.parent.mkdir(parents=True, exist_ok=True)
        data = self.get_simulation_configuration_data()

        with output_filename.open("w", encoding="utf-8") as file:
            file.write(f"# configuration_{self.software}\n")
            file.write("\n\n")
            if self.software == "sim_telarray":
                data.sort(key=lambda x: (x[0], x[1]))
                for telescope, group in groupby(data, key=lambda x: x[0]):
                    file.write(f"## [{telescope}]({telescope}.md)\n")
                    file.write("\n\n")
                    self._write_to_file(group, file)
            else:
                self._write_to_file(data, file)

    def produce_array_element_report(self):
        """
        Produce a markdown report of all model parameters per array element.

        Outputs one markdown report of a given array element listing parameter values,
        versions, and descriptions.
        """
        if self.observatory:
            self.produce_observatory_report()
            return

        telescope_model = TelescopeModel(
            site=self.site,
            telescope_name=self.array_element,
            model_version=self.model_version,
            label="reports",
            db_config=self.db_config,
            ignore_software_version=True,
        )

        output_filename = Path(self.output_path / (telescope_model.name + ".md"))
        output_filename.parent.mkdir(parents=True, exist_ok=True)
        data = self.get_array_element_parameter_data(telescope_model)
        # Sort data by class to prepare for grouping
        if not isinstance(data, str):
            data.sort(key=lambda x: (x[0], x[1]), reverse=True)

        with output_filename.open("w", encoding="utf-8") as file:
            # Group by class and write sections
            file.write(f"# {telescope_model.name}\n")

            if telescope_model.name != telescope_model.design_model:
                file.write(
                    "The design model can be found here: "
                    f"[{telescope_model.design_model}]"
                    f"({telescope_model.design_model}.md).\n\n"
                )
                file.write(
                    "Parameters shown in ***bold and italics*** are specific to each telescope.\n"
                    "Parameters without emphasis are inherited from the design model.\n"
                )
                file.write("\n\n")

            for class_name, group in groupby(data, key=lambda x: x[0]):
                group = sorted(group, key=lambda x: x[1])
                file.write(f"## {class_name}\n\n")
                self._write_to_file(group, file)

    def produce_model_parameter_reports(self, collection="telescopes"):
        """
        Produce a markdown report per parameter for a given array element.

        Outputs one markdown report per model parameter of a given array element comparing
        values across model versions.
        """
        logger.info(
            f"Comparing parameters across model versions for Telescope: {self.array_element}"
            f" and Site: {self.site}."
        )
        output_path = self.output_path / f"{self.array_element}"
        Path(output_path).mkdir(parents=True, exist_ok=True)

        all_parameter_names = names.model_parameters(None).keys()
        all_parameter_data = self.db.get_model_parameters_for_all_model_versions(
            site=self.site, array_element_name=self.array_element, collection=collection
        )

        comparison_data = self._compare_parameter_across_versions(
            all_parameter_data, all_parameter_names
        )

        for parameter in all_parameter_names:
            parameter_data = comparison_data.get(parameter)
            if not parameter_data:
                continue

            output_filename = output_path / f"{parameter}.md"

            parameter_descriptions = self.get_all_parameter_descriptions(collection=collection).get(
                parameter
            ) or self.get_all_parameter_descriptions(collection="telescopes").get(parameter)

            description = parameter_descriptions.get("description")
            with output_filename.open("w", encoding="utf-8") as file:
                # Write header and table
                self._write_parameter_header(file, parameter, description)
                self._write_table_rows(file, parameter, comparison_data)

                # If entries reference files, write the image/plot section
                if comparison_data.get(parameter)[0]["file_flag"]:
                    self._write_file_flag_section(file, parameter, comparison_data)

    def _write_parameter_header(self, file, parameter, description):
        """Write the markdown header for a parameter file."""
        file.write(
            f"# {parameter}\n\n"
            f"**Telescope**: {self.array_element}\n\n"
            f"**Description**: {description}\n\n"
            "\n"
        )

    def _write_table_rows(self, file, parameter, comparison_data):
        """Write the comparison table header and rows for a parameter."""
        # Write table header
        file.write(
            "| Parameter Version      | Model Version(s)      "
            "| Value                |\n"
            "|------------------------|--------------------"
            "|----------------------|\n"
        )

        # Write table rows
        for item in comparison_data.get(parameter):
            file.write(
                f"| {item['parameter_version']} | {item['model_version']} | {item['value']} |\n"
            )

        file.write("\n")

    def _write_file_flag_section(self, file, parameter, comparison_data):
        """Write image/plot references when parameter entries include files."""
        outpath = Path(io_handler.IOHandler().get_output_directory().parent / "_images")
        latest_parameter_version = max(
            comparison_data.get(parameter),
            key=lambda x: tuple(map(int, x["parameter_version"].split("."))),
        )["parameter_version"]

        all_model_versions = []
        for item in comparison_data.get(parameter):
            model_versions = item["model_version"].split(", ")
            all_model_versions.extend(model_versions)

        latest_model_version = max(all_model_versions, key=lambda x: tuple(map(int, x.split("."))))
        tel = self._get_telescope_identifier(latest_model_version)

        file.write("The latest parameter version is plotted below.\n\n")

        if parameter != "camera_config_file":
            plot_name = f"{parameter}_{latest_parameter_version}_{self.site}_{tel}"
            image_path = outpath / f"{plot_name}.png"
            file.write(f"![Parameter plot.]({image_path.as_posix()})")
            return

        # camera_config_file: find latest value and convert markdown link to png filename
        latest_value = None
        for item in comparison_data.get(parameter):
            if latest_model_version in item["model_version"].split(", "):
                latest_value = item["value"]
                break

        if latest_value is None:
            return

        match = MARKDOWN_LINK_RE.search(latest_value)
        if not match:
            return

        filename_png = Path(match.group(1)).with_suffix(".png").name
        image_path = outpath / filename_png
        file.write(f"![Camera configuration plot.]({image_path.as_posix()})")

    def _write_array_layouts_section(self, file, layouts):
        """Write the array layouts section of the report."""
        file.write("\n## Array Layouts\n\n")
        for layout in layouts:
            layout_name = layout["name"]
            elements = layout["elements"]
            file.write(f"### {layout_name}\n\n")
            file.write("| Element |\n|---------|\n")
            for element in sorted(elements):
                file.write(f"| [{element}]({element}.md) |\n")
            file.write("\n")
            version = self.model_version.replace(".", "-")
            filename = f"OBS-{self.site}_{layout_name}_{version}.png"
            image_path = f"/_images/{filename}"
            file.write(f"![{layout_name} Layout]({image_path})\n\n")
            file.write("\n")

    def _write_array_triggers_section(self, file, trigger_configs):
        """Write the array triggers section of the report."""
        file.write("\n## Array Trigger Configurations\n\n")
        file.write(
            "| Trigger Name | Multiplicity | Width | Hard Stereo | Min Separation |\n"
            "|--------------|--------------|--------|-------------|----------------|\n"
        )
        for config in trigger_configs:
            name = config["name"]
            mult = f"{config['multiplicity']['value']} {config['multiplicity']['unit'] or ''}"
            width = f"{config['width']['value']} {config['width']['unit'] or ''}"
            stereo = "Yes" if config["hard_stereo"]["value"] else "No"
            min_sep = (
                f"{config['min_separation']['value']} {config['min_separation']['unit'] or '-'}"
            )
            file.write(
                f"| {name} | {mult.strip()} | {width.strip()} | {stereo} | {min_sep.strip()} |\n"
            )
        file.write("\n")

    def _write_parameters_table(self, file, all_parameter_data):
        """Write the main parameters table of the report."""
        file.write(
            "| Parameter | Value | Parameter Version |\n"
            "|-----------|--------|-------------------|\n"
        )
        for param_name, param_data in sorted(all_parameter_data.items()):
            value = param_data.get("value")
            unit = param_data.get("unit") or " "
            file_flag = param_data.get("file", False)
            parameter_version = param_data.get("parameter_version")

            if value is None:
                continue

            if param_name == "array_layouts":
                file.write(
                    "| array_layouts | [View Array Layouts](#array-layouts)"
                    f" | {parameter_version} |\n"
                )
            elif param_name == "array_triggers":
                file.write(
                    "| array_triggers | [View Trigger Configurations]"
                    f"(#array-trigger-configurations) | {parameter_version} |\n"
                )
            else:
                formatted_value = self._format_parameter_value(
                    param_name, value, unit, file_flag, parameter_version
                )
                file.write(f"| {param_name} | {formatted_value} | {parameter_version} |\n")
        file.write("\n")

    def produce_observatory_report(self):
        """Produce a markdown report of all observatory parameters for a given site."""
        output_filename = Path(self.output_path / f"OBS-{self.site}.md")
        output_filename.parent.mkdir(parents=True, exist_ok=True)

        all_parameter_data = self.db.get_model_parameters(
            site=self.site,
            array_element_name="OBS-" + self.site,
            collection="sites",
            model_version=self.model_version,
        )

        if not all_parameter_data:
            logger.warning(f"No observatory parameters found for site {self.site}")
            return

        Path(f"{self.output_path}/model").mkdir(parents=True, exist_ok=True)
        self.db.export_model_files(parameters=all_parameter_data, dest=f"{self.output_path}/model")

        with output_filename.open("w", encoding="utf-8") as file:
            file.write(f"# Observatory Parameters - {self.site} Site\n\n")
            self._write_parameters_table(file, all_parameter_data)

            if "array_layouts" in all_parameter_data:
                self._write_array_layouts_section(
                    file, all_parameter_data["array_layouts"]["value"]
                )

            if "array_triggers" in all_parameter_data:
                self._write_array_triggers_section(
                    file, all_parameter_data["array_triggers"]["value"]
                )

    def get_calibration_data(self, all_parameter_data, array_element):
        """Get calibration data and descriptions for a given array element."""
        calibration_descriptions = self.get_all_parameter_descriptions(
            collection="calibration_devices"
        )
        # get descriptions of array element positions from the telescope collection
        telescope_descriptions = self.get_all_parameter_descriptions(collection="telescopes")
        data = []
        class_grouped_data = {}

        for parameter in all_parameter_data.keys():
            parameter_descriptions = calibration_descriptions.get(
                parameter
            ) or telescope_descriptions.get(parameter)

            parameter_data = all_parameter_data.get(parameter)
            parameter_version = parameter_data.get("parameter_version")
            unit = parameter_data.get("unit") or " "
            value_data = parameter_data.get("value")

            if value_data is None:
                continue

            file_flag = parameter_data.get("file", False)
            value = self._format_parameter_value(
                parameter, value_data, unit, file_flag, parameter_version
            )

            description = parameter_descriptions.get("description")
            short_description = parameter_descriptions.get("short_description") or description

            inst_class = parameter_descriptions.get("inst_class")

            matching_instrument = parameter_data["instrument"] == array_element
            if not names.is_design_type(array_element) and matching_instrument:
                parameter = f"***{parameter}***"
                parameter_version = f"***{parameter_version}***"
                if not self.is_markdown_link(value):
                    value = f"***{value}***"
                description = f"***{description}***"
                short_description = f"***{short_description}***"

            # Group by class name
            if inst_class not in class_grouped_data:
                class_grouped_data[inst_class] = []

            class_grouped_data[inst_class].append(
                [
                    inst_class,
                    parameter,
                    parameter_version,
                    value,
                    description,
                    short_description,
                ]
            )

        data = []
        for class_name in sorted(class_grouped_data.keys(), reverse=True):
            sorted_class_data = sorted(class_grouped_data[class_name], key=lambda x: x[1])
            data.extend(sorted_class_data)

        return data

    def is_markdown_link(self, value):
        """
        Return True if the string is a Markdown-style link: [text](target).

        Parameters
        ----------
        value : str
            The string to check.

        Returns
        -------
        bool
            True if the string is a Markdown link, False otherwise.
        """
        return bool(re.fullmatch(r"\[[^\]]*\]\([^)]+\)", value.strip()))

    def produce_calibration_reports(self):
        """Write calibration reports."""
        array_elements = self._collect_calibration_array_elements()

        for calibration_device in array_elements:
            device_sites = names.get_site_from_array_element_name(calibration_device)
            site = device_sites[0] if isinstance(device_sites, list) else device_sites

            all_parameter_data = self.db.get_model_parameters(
                site=site,
                array_element_name=calibration_device,
                collection="calibration_devices",
                model_version=self.model_version,
            )

            output_filename = Path(self.output_path / (f"{calibration_device}.md"))
            output_filename.parent.mkdir(parents=True, exist_ok=True)
            data = self.get_calibration_data(all_parameter_data, calibration_device)

            design_model = self.db.get_design_model(
                self.model_version, calibration_device, "calibration_devices"
            )

            self._write_single_calibration_report(
                output_filename, calibration_device, data, design_model
            )

    def _collect_calibration_array_elements(self):
        """Return a list of calibration devices including their design models."""
        calibration_array_elements = self.db.get_array_elements(
            self.model_version, collection="calibration_devices"
        )
        array_elements = calibration_array_elements.copy()
        for element in calibration_array_elements:
            design_model = self.db.get_design_model(
                self.model_version, element, "calibration_devices"
            )
            if design_model and design_model not in array_elements:
                array_elements.append(design_model)
        return array_elements

    def _write_single_calibration_report(
        self, output_filename, calibration_device, data, design_model
    ):
        """Write a single calibration device markdown report file."""
        with output_filename.open("w", encoding="utf-8") as file:
            file.write(f"# {calibration_device}\n")
            file.write("\n\n")

            if not names.is_design_type(calibration_device):
                file.write(
                    f"The design model can be found here: [{design_model}]({design_model}.md).\n\n"
                )
                file.write(
                    "Parameters shown in ***bold and italics*** are specific"
                    " to each array element.\n"
                    "Parameters without emphasis are inherited from the design model.\n"
                )
                file.write("\n\n")

            for class_name, group in groupby(data, key=lambda x: x[0]):
                group = sorted(group, key=lambda x: x[1])
                file.write(f"## {class_name}\n\n")
                # Transform parameter display names for human-readable report
                display_group = []
                for row in group:
                    param = row[1]
                    # Preserve markdown emphasis (***param***) if present
                    if isinstance(param, str) and param.startswith("***") and param.endswith("***"):
                        inner = param.strip("*")
                        new_inner = inner.replace("_", " ")
                        new_param = f"***{new_inner}***"
                    elif isinstance(param, str):
                        new_param = param.replace("_", " ")
                    else:
                        new_param = param

                    new_row = [row[0], new_param, row[2], row[3], row[4], row[5]]
                    display_group.append(new_row)

                self._write_to_file(display_group, file)

    def generate_model_parameter_reports_for_devices(self, array_elements):
        """Create model-parameter comparison reports for calibration devices."""
        for calibration_device in array_elements:
            device_sites = names.get_site_from_array_element_name(calibration_device)
            # parameters are site independent so just take the first site to read from db
            site = device_sites[0] if isinstance(device_sites, list) else device_sites
            self.site = site
            self.array_element = calibration_device
            self.produce_model_parameter_reports(collection="calibration_devices")



#!/usr/bin/python3

r"""
    Calculate the camera FoV of the telescope requested and plot the camera.

    The orientation for the plotting is "as seen for an observer facing the camera".

    An example of the camera plot can be found below.

    .. _camera_fov_plot:
    .. image:: images/validate_camera_fov_North-LST-1_pixelLayout.png
      :width: 50 %


    Command line arguments
    ----------------------
    site (str, required)
        North or South.
    telescope (str, required)
        Telescope model name (e.g. LST-1, SST-D, ...)
    model_version (str, optional)
        Model version
    camera_in_sky_coor (bool, optional)
        Plot the camera layout in sky coordinates akin to looking at it from behind for single \
         mirror telescopes
    print_pixels_id (bool, optional)
        Up to which pixel ID to print. To suppress printing of pixel IDs, set to zero\
         (--print_pixels_id 0). To print all pixels, set to 'All'."

    Example
    -------
    LST - 5.0.0

    .. code-block:: console

        simtools-validate-camera-fov --site North \\
            --telescope LSTN-01 --model_version 5.0.0

    The output is saved in simtools-output/validate_camera_fov.

    Expected final print-out message:

    .. code-block:: console

        Saved camera plot in /workdir/external/simtools/simtools-output/validate_camera_fov\\
        /application-plots/validate_camera_fov_LST-1_pixel_layout.png

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model.telescope_model import TelescopeModel
from simtools.visualization import plot_camera, visualize


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Calculate the camera FoV of the telescope requested. "
            "Plot the camera, as seen for an observer facing the camera."
        ),
    )
    config.parser.add_argument(
        "--camera_in_sky_coor",
        help=(
            "Plot the camera layout in sky coordinates "
            "(akin to looking at it from behind for single mirror telescopes)"
        ),
        action="store_true",
        default=False,
    )
    config.parser.add_argument(
        "--print_pixels_id",
        help=(
            "Up to which pixel ID to print. "
            "To suppress printing of pixel IDs, set to zero (--print_pixels_id 0). "
            "To print all pixels, set to 'All'."
        ),
        default=50,
    )
    return config.initialize(db_config=True, simulation_model=["telescope", "model_version"])


def main():
    """Validate camera field of view."""
    app_context = startup_application(_parse)

    label = "validate_camera_fov"

    tel_model = TelescopeModel(
        site=app_context.args["site"],
        telescope_name=app_context.args["telescope"],
        db_config=app_context.db_config,
        model_version=app_context.args["model_version"],
        label=label,
    )
    tel_model.export_model_files()

    app_context.logger.info(f"\nValidating the camera FoV of {tel_model.name}\n")

    focal_length = tel_model.get_telescope_effective_focal_length("cm")
    camera = tel_model.camera

    fov, r_edge_avg = camera.calc_fov()

    app_context.logger.info(f"\nEffective focal length = {focal_length:.3f} cm")
    app_context.logger.info(f"{tel_model.name} FoV = {fov:.3f} deg")
    app_context.logger.info(f"Avg. edge radius = {r_edge_avg:.3f} cm\n")

    # Now plot the camera as well
    try:
        pixel_ids_to_print = int(app_context.args["print_pixels_id"])
        if pixel_ids_to_print == 0:
            pixel_ids_to_print = -1  # so not print the zero pixel
    except ValueError as exc:
        if app_context.args["print_pixels_id"].lower() == "all":
            pixel_ids_to_print = camera.get_number_of_pixels()
        else:
            raise ValueError(
                f"The value provided to --print_pixels_id ({app_context.args['print_pixels_id']}) "
                "should be an integer or All"
            ) from exc
    fig = plot_camera.plot_pixel_layout(
        camera, app_context.args["camera_in_sky_coor"], pixel_ids_to_print
    )
    output_dir = app_context.io_handler.get_output_directory()
    plot_file_prefix = output_dir.joinpath(f"{label}_{tel_model.name}_pixel_layout")
    visualize.save_figure(fig, f"{plot_file_prefix!s}", log_title="camera")


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
Simulate illuminator (distant calibration light source).

Illuminators are calibration light sources not attached to a particular telescope.
Two types of illuminators are supported:

1. Illuminator as foreseen at CTAO with fixed positions as defined in the simulation models
   database.
2. Illuminator at a configurable position relative to the array center. Note that in this case
   the telescope pointing is fixed towards zenith.

Example Usage
-------------

1. Simulate illuminator with positions as defined in the simulation models database:

    .. code-block:: console

        simtools-simulate-illuminator --light_source ILLN-01 \
        --telescope MSTN-04 --site North \
        --model_version 6.0.0

2. Simulate at a configurable position (1km above array center) and pointing downwards:

    .. code-block:: console

        simtools-simulate-illuminator --light_source ILLN-01 \
        --light_source_position 0. 0. 1000. \
        --light_source_pointing 0. 0. -1. \
        --telescope MSTN-15 --site North \
        --model_version 6.0.0

Command Line Arguments
----------------------
light_source (str, optional)
    Illuminator in array, e.g., ILLN-01.
number_of_events (int, optional)
    Number of events to simulate.
telescope (str, required)
    Telescope model name (e.g. LSTN-01, SSTS-design, SSTS-25, ...)
site (str, required)
    Site name (North or South).
model_version (str, optional)
    Version of the simulation model.
light_source_position (float, float, float, optional)
    Light source position (x,y,z) relative to the array center (ground coordinates) in
    m. If not set, the position from the simulation model is used.
light_source_pointing (float, float, float, optional)
    Light source pointing direction. If not set, the pointing from the simulation model is used.
output_prefix (str, optional)
    Prefix for output files (default: empty).
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.simtel.simulator_light_emission import SimulatorLightEmission


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Simulate light emission by a calibration light source (not attached to a telescope)."
        ),
    )
    config.parser.add_argument(
        "--light_source",
        help="Illuminator name, i.e. ILLN-design",
        type=str,
        default=None,
        required=True,
    )
    configurable_light_source_args = config.parser.add_argument_group(
        "Configurable light source position and pointing (override simulation model values)"
    )
    configurable_light_source_args.add_argument(
        "--light_source_position",
        help="Light source position (x,y,z) relative to the array center (ground coordinates) in m",
        metavar=("X", "Y", "Z"),
        nargs=3,
        required=False,
    )
    configurable_light_source_args.add_argument(
        "--light_source_pointing",
        help=(
            "Light source pointing direction "
            "(Example for pointing downwards: --light_source_pointing 0 0 -1)"
        ),
        metavar=("X", "Y", "Z"),
        nargs=3,
        required=False,
    )
    config.parser.add_argument(
        "--number_of_events",
        help="Number of events to simulate",
        type=int,
        default=1,
        required=False,
    )
    config.parser.add_argument(
        "--output_prefix",
        help="Prefix for output files (default: empty)",
        type=str,
        default=None,
        required=False,
    )
    return config.initialize(
        db_config=True,
        simulation_model=["telescope", "model_version"],
        require_command_line=True,
    )


def main():
    """Simulate light emission from illuminator."""
    app_context = startup_application(_parse)

    light_source = SimulatorLightEmission(
        light_emission_config=app_context.args,
        db_config=app_context.db_config,
        label=app_context.args.get("label"),
    )
    light_source.simulate()


if __name__ == "__main__":
    main()



"""Utilities for managing the simulation models repository.

Simulation model parameters and production tables are managed through
a gitlab repository ('simulation_models'). This module provides service
functions to interact with and verify the repository.

Main functionalities are:

- validation of production tables against model parameters
- generation of new production tables and model parameters based on
  updates defined in a configuration file

"""

import logging
from pathlib import Path

from packaging.version import Version
from packaging.version import parse as parse_version

import simtools.data_model.model_data_writer as writer
from simtools.io import ascii_handler
from simtools.utils import names

_logger = logging.getLogger(__name__)


def get_production_directory(simulation_models_path, model_version=None):
    """
    Get the production directory for a specific model version.

    Parameters
    ----------
    simulation_models_path : str
        Path to the simulation models repository.
    model_version : str, optional
        Specific model version to get the production directory for.

    Returns
    -------
    Path
        Path to the production directory.
    """
    if model_version:
        return Path(simulation_models_path) / "simulation-models/productions" / str(model_version)
    return Path(simulation_models_path) / "simulation-models/productions"


def get_model_parameter_directory(simulation_models_path):
    """
    Get the model parameters directory.

    Parameters
    ----------
    simulation_models_path : str
        Path to the simulation models repository.

    Returns
    -------
    Path
        Path to the model parameters directory.
    """
    return Path(simulation_models_path) / "simulation-models/model_parameters"


def verify_simulation_model_production_tables(simulation_models_path):
    """
    Verify the simulation model production tables in the specified path.

    Checks that all model parameters defined in the production tables are
    present in the simulation models repository.

    Parameters
    ----------
    simulation_models_path : str
        Path to the simulation models repository.

    Returns
    -------
    bool
        True if all parameters found, False if any missing.
    """
    productions_path = get_production_directory(simulation_models_path)
    production_files = list(productions_path.rglob("*.json"))

    _logger.info(
        f"Verifying {len(production_files)} simulation model production "
        f"tables in {simulation_models_path}"
    )

    missing_files = []
    total_checked = 0

    for production_file in production_files:
        file_missing, file_checked = _verify_model_parameters_for_production(
            simulation_models_path, production_file
        )
        missing_files.extend(file_missing)
        total_checked += file_checked

    _logger.info(f"Checked {total_checked} parameters, {len(missing_files)} missing")

    if missing_files:
        for missing_file in missing_files:
            _logger.error(f"Missing: {missing_file}")
        return False

    _logger.info("Verification passed: All parameters found")
    return True


def _verify_model_parameters_for_production(simulation_models_path, production_file):
    """
    Verify that model parameters defined in the production tables exist.

    Parameters
    ----------
    simulation_models_path : str
        Path to the simulation models repository.
    production_file : Path
        Path to the production file.

    Returns
    -------
    tuple
        (missing_files_list, total_checked_count)
    """
    production_table = ascii_handler.collect_data_from_file(production_file)
    missing_files = []
    total_checked = 0

    parameters = production_table.get("parameters", {})
    for array_element, par_dict in parameters.items():
        if isinstance(par_dict, dict):
            for param_name, param_version in par_dict.items():
                total_checked += 1
                parameter_file = _get_model_parameter_file_path(
                    simulation_models_path, array_element, param_name, param_version
                )
                if parameter_file and not parameter_file.exists():
                    missing_files.append(str(parameter_file))

    return missing_files, total_checked


def _get_model_parameter_file_path(
    simulation_models_path, array_element, parameter_name, parameter_version
):
    """
    Get the file path for a model parameter.

    Take into account path structure based on collections and array elements.

    Parameters
    ----------
    simulation_models_path : str
        Path to the simulation models repository.
    array_element : str
        Name of the array element (e.g., 'telescope').
    parameter_name : str
        Name of the parameter.
    parameter_version : str
        Version of the parameter.

    Returns
    -------
    Path
        The file path to the model parameter JSON file.
    """
    collection = names.get_collection_name_from_parameter_name(parameter_name)
    return (
        get_model_parameter_directory(simulation_models_path)
        / (
            collection
            if collection in ("configuration_sim_telarray", "configuration_corsika")
            else ""
        )
        / (array_element if collection != "configuration_corsika" else "")
        / parameter_name
        / f"{parameter_name}-{parameter_version}.json"
    )


def generate_new_production(model_version, simulation_models_path):
    """
    Generate a new production definition (production tables and model parameters).

    The following steps are performed:

    - copy of production tables from an existing base model version
    - update production tables with changes defined in a configuration file (expected
      to be called 'info.yml' in the target production directory)
    - generate new model parameter entries for changed parameters
    - allows for full or patch updates

    Parameters
    ----------
    model_version: str
        Model version to be created or updated.
    simulation_models_path: str
        Path to the simulation models repository.
    """
    modification_dict = _get_changes_dict(model_version, simulation_models_path)
    update_type = modification_dict.get("model_update", "full_update")
    changes, base_model_version = _get_changes_to_production(
        modification_dict, simulation_models_path, update_type
    )

    _apply_changes_to_production_tables(
        changes,
        base_model_version,
        modification_dict["model_version"],
        update_type,
        simulation_models_path,
    )

    _apply_changes_to_model_parameters(changes, simulation_models_path)


def _apply_changes_to_production_tables(
    changes, base_model_version, model_version, update_type, simulation_models_path
):
    """
    Apply changes to or generate new production tables and write them to target directory.

    Parameters
    ----------
    changes: dict
        Changes to be applied.
    base_model_version: str
        Base model version (source directory for production tables).
    model_version: str
        Model version of the new production tables.
    update_type: str
        Update type (e.g., 'full_update' or 'patch_update').
    simulation_models_path: Path
        Path to the simulation models repository.
    """
    source = get_production_directory(simulation_models_path, base_model_version)
    target = get_production_directory(simulation_models_path, model_version)
    _logger.info(f"Production tables {update_type} from {source} to {target}")
    target.mkdir(parents=True, exist_ok=True)

    # load existing tables from source
    tables = {}
    for file_path in Path(source).rglob("*.json"):
        data = ascii_handler.collect_data_from_file(file_path)
        if not isinstance(data, dict):
            raise TypeError(f"Unsupported data type {type(data)} in {file_path}")
        tables[data["production_table_name"]] = data

    # placeholder for new tables
    for table_name in changes:
        tables.setdefault(table_name, {})

    for table_name, data in tables.items():
        if _apply_changes_to_production_table(
            table_name, data, changes, model_version, update_type == "patch_update"
        ):
            target_file = target / f"{table_name}.json"
            _logger.info(f"Writing updated production table '{target_file}'")
            data["production_table_name"] = table_name
            ascii_handler.write_data_to_file(data, target_file, sort_keys=True)


def _apply_changes_to_production_table(table_name, data, changes, model_version, patch_update):
    """
    Apply changes to a single production table.

    Parameters
    ----------
    table_name: str
        Name of the production table.
    data: dict
        Data to be updated.
    changes: dict
        Changes to be applied.
    model_version: str
        Model version of the new production tables.
    patch_update: bool
        True if patch update (modify only changed parameters), False for full update.

    Returns
    -------
    bool
        True if data was modified and should be written to disk (patch updates);
        always True for full updates.
    """
    data["model_version"] = model_version
    if table_name in changes:
        table_parameters = {} if patch_update else data.get("parameters", {}).get(table_name, {})
        parameters, deprecated = _update_parameters_dict(table_parameters, changes, table_name)
        data["parameters"] = parameters
        if deprecated and patch_update:
            data["deprecated_parameters"] = deprecated
    elif patch_update:
        return False

    return True


def _get_changes_dict(model_version, simulation_models_path):
    """
    Load the changes dictionary from 'info.yml' files in production directories.

    Parameters
    ----------
    model_version: str
        Model version of the new production tables.
    simulation_models_path: Path
        Path to the simulation models directory.

    Returns
    -------
    dict
        Changes dictionary.
    """
    return ascii_handler.collect_data_from_file(
        get_production_directory(simulation_models_path, model_version) / "info.yml"
    )


def _get_changes_to_production(
    modification_dict, simulation_models_path, update_type="full_update"
):
    """
    Prepare changes applied to production tables.

    For full updates, this includes the combination of changes to be applied
    for all model versions in the history, starting from the base version.

    Parameters
    ----------
    modification_dict: dict
        Modifications dictionary.
    simulation_models_path: Path
        Path to the simulation models directory.
    update_type: str
        Update mode.

    Returns
    -------
    dict, str
        Changes dictionary and base model version.
    """
    model_version_history = modification_dict.get("model_version_history", [])

    try:
        # oldest version is the base version
        base_model_version = min(set(model_version_history), key=Version)
    except ValueError:
        _logger.debug(f"Base model version not found in {model_version_history}")
        return {}, modification_dict.get("model_version")

    changes = modification_dict.get("changes", {})
    if update_type == "patch_update":
        return changes, base_model_version

    for version_mod in reversed(model_version_history):
        _changes_dict = _get_changes_dict(version_mod, simulation_models_path)
        _version_changes, base_model_version = _get_changes_to_production(
            _changes_dict, simulation_models_path, update_type="full_update"
        )
        changes = _update_two_levels_in_changes_dict(changes, _version_changes)
        # stop iterative loop after reaching first full version of production tables
        if _changes_dict.get("model_update", "full_update") == "full_update":
            break

    return changes, base_model_version


def _update_two_levels_in_changes_dict(d, u):
    """Update changes dict, e.g. {"LSTN-design": { "parameter_name: { ... } } }."""
    for k, v in u.items():
        if isinstance(v, dict) and isinstance(d.get(k), dict):
            d[k].update(v)
        else:
            d[k] = v
    return d


def _update_parameters_dict(table_parameters, changes, table_name):
    """
    Create a new parameters dictionary for the production tables.

    Include only changes relevant to the specific telescope.
    Do not include parameters if 'deprecated' flag is set to True.

    Parameters
    ----------
    table_parameters: dict
        Parameters for the specific table.
    changes: dict
        The changes to be applied, containing table and parameter information.
    table_name: str
        The name of the production table to filter parameters for.

    Returns
    -------
    dict, list
        Dictionary containing only the new/changed parameters for the specified table.
        List of deprecated parameters.
    """
    new_params = {table_name: table_parameters}
    deprecated_params = []

    for param, data in changes[table_name].items():
        if data.get("deprecated", False):
            _logger.info(f"Removing model parameter '{table_name} - {param}'")
            deprecated_params.append(param)
            new_params[table_name].pop(param, None)
        else:
            version = data["version"]
            _logger.info(f"Setting '{table_name} - {param}' to version {version}")
            new_params[table_name][param] = version

    return new_params, deprecated_params


def _apply_changes_to_model_parameters(changes, simulation_models_path):
    """
    Apply changes to model parameters by creating new parameter entries.

    Parameters
    ----------
    changes: dict
        The changes to be applied.
    simulation_models_path: Path
        Path to the simulation models directory.
    """
    for telescope, parameters in changes.items():
        for param, param_data in parameters.items():
            if param_data.get("value") is not None:
                _create_new_model_parameter_entry(
                    telescope, param, param_data, simulation_models_path
                )


def _create_new_model_parameter_entry(telescope, param, param_data, simulation_models_path):
    """
    Create new model parameter entry in the model parameters directory.

    If a model parameter files exists, copy latest version and update the fields.
    Otherwise generate new file using the model parameter schema.

    Parameters
    ----------
    telescope: str
        Name of the telescope.
    param: str
        Name of the parameter.
    param_data: dict
        Dictionary containing the parameter data including version and value.
    simulation_models_path: Path
        Path to the simulation models directory.
    """
    telescope_dir = get_model_parameter_directory(simulation_models_path) / telescope
    if not telescope_dir.exists():
        _logger.info(f"Create directory for array element '{telescope}': '{telescope_dir}'.")
        telescope_dir.mkdir(parents=True, exist_ok=True)

    param_dir = telescope_dir / param
    try:
        latest_file = _get_latest_model_parameter_file(param_dir, param)
    except FileNotFoundError:
        latest_file = None

    if latest_file is not None:
        json_data = ascii_handler.collect_data_from_file(latest_file)
        param_data["version"] = _check_for_major_version_jump(
            json_data, param_data, param, telescope
        )
        # important for e.g. nsb_pixel_rate
        if isinstance(json_data["value"], list) and not isinstance(param_data["value"], list):
            param_data["value"] = [param_data["value"]] * len(json_data["value"])
        param_data["meta_parameter"] = json_data.get("meta_parameter", False)

    writer.ModelDataWriter.dump_model_parameter(
        parameter_name=param,
        value=param_data["value"],
        instrument=telescope,
        parameter_version=param_data["version"],
        output_file=f"{param}-{param_data['version']}.json",
        output_path=param_dir,
        unit=param_data.get("unit"),
        meta_parameter=param_data.get("meta_parameter", False),
    )


def _get_latest_model_parameter_file(directory, parameter):
    """
    Get the latest model parameter JSON file for a parameter in the given directory.

    Assume files are named in the format 'parameter-version.json'.

    Parameters
    ----------
    directory: str
        Path to the directory containing parameter JSON files.
    parameter: str
        Name of the parameter to find.

    Returns
    -------
    str
        Path to the latest JSON file for the parameter.

    Raises
    ------
    FileNotFoundError
        If no files for the parameter are found in the directory.
    """
    directory_path = Path(directory)
    files = list(directory_path.glob(f"{parameter}-*.json"))
    if not files:
        raise FileNotFoundError(
            f"No JSON files found for parameter '{parameter}' in directory '{directory}'."
        )

    def extract_version(path: Path):
        # version is part after first '-'
        return parse_version(path.stem.split("-", 1)[1])

    latest_file = max(files, key=extract_version)
    return str(latest_file)


def _check_for_major_version_jump(json_data, param_data, param, telescope):
    """
    Check for major version jump and print a warning if necessary.

    Generally a jump from e.g. '3.1.0' to '5.0.0' should be avoided.
    """
    latest_version = parse_version(json_data.get("parameter_version", "0"))
    new_version = parse_version(param_data["version"])
    if new_version.major > latest_version.major + 1:
        _logger.warning(
            f"Major version jump from {latest_version} to {new_version} "
            f"for parameter '{param}' in telescope '{telescope}'."
        )
    return param_data["version"]



"""
Simtools dependencies version management.

This modules provides two main functionalities:

- retrieve the versions of simtools dependencies (e.g., databases, sim_telarray, CORSIKA)
- provide space for future implementations of version management

"""

import logging
import os
import re
import subprocess
from pathlib import Path

import yaml

from simtools.io import ascii_handler
from simtools.version import __version__

_logger = logging.getLogger(__name__)


def get_version_string(db_config=None, run_time=None):
    """
    Print the versions of the dependencies.

    Parameters
    ----------
    db_config : dict, optional
        Database configuration dictionary.
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    str
        String containing the versions of the dependencies.

    """
    return (
        f"Database name: {get_database_version_or_name(db_config, version=False)}\n"
        f"Database version: {get_database_version_or_name(db_config, version=True)}\n"
        f"sim_telarray version: {get_sim_telarray_version(run_time)}\n"
        f"CORSIKA version: {get_corsika_version(run_time)}\n"
        f"Build options: {get_build_options(run_time)}\n"
        f"Runtime environment: {run_time if run_time else 'None'}\n"
    )


def get_software_version(software):
    """
    Return the version of the specified software package.

    Parameters
    ----------
    software : str
        Name of the software package.

    Returns
    -------
    str
        Version of the specified software package.
    """
    if software.lower() == "simtools":
        return __version__

    try:
        version_call = f"get_{software.lower()}_version"
        return globals()[version_call]()
    except KeyError as exc:
        raise ValueError(f"Unknown software: {software}") from exc


def get_database_version_or_name(db_config, version=True):
    """
    Get the version or name of the simulation model data base used.

    Parameters
    ----------
    db_config : dict
        Dictionary containing the database configuration.
    version : bool
        If True, return the version of the database. If False, return the name.

    Returns
    -------
    str
        Version or name of the simulation model data base used.

    """
    if version:
        return db_config and db_config.get("db_simulation_model_version")
    return db_config and db_config.get("db_simulation_model")


def get_sim_telarray_version(run_time=None):
    """
    Get the version of the sim_telarray package using 'sim_telarray --version'.

    Version strings for sim_telarray are of the form "2024.271.0" (year.day_of_year.patch).

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    str
        Version of the sim_telarray package.
    """
    sim_telarray_path = os.getenv("SIMTOOLS_SIMTEL_PATH")
    if sim_telarray_path is None:
        _logger.warning("Environment variable SIMTOOLS_SIMTEL_PATH is not set.")
        return None
    sim_telarray_path = Path(sim_telarray_path) / "sim_telarray" / "bin" / "sim_telarray"

    if run_time is None:
        command = [str(sim_telarray_path), "--version"]
    else:
        command = [*run_time, str(sim_telarray_path), "--version"]

    _logger.debug(f"Running command: {command}")
    result = subprocess.run(command, capture_output=True, text=True, check=False)

    # expect stdout with e.g. a line 'Release: 2024.271.0 from 2024-09-27'
    match = re.search(r"^Release:\s+(.+)", result.stdout, re.MULTILINE)
    if match:
        return match.group(1).split()[0]

    _logger.debug(f"Command output stdout: {result.stdout} stderr: {result.stderr}")

    raise ValueError(f"sim_telarray release not found in {result.stdout}")


def get_corsika_version(run_time=None):
    """
    Get the version of the CORSIKA package.

    Version strings for CORSIKA are of the form "7.7550" (major.minor with 4-digit minor).

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    str
        Version of the CORSIKA package.
    """
    version = None
    sim_telarray_path = os.getenv("SIMTOOLS_SIMTEL_PATH")
    if sim_telarray_path is None:
        _logger.warning("Environment variable SIMTOOLS_SIMTEL_PATH is not set.")
        return None
    corsika_command = Path(sim_telarray_path) / "corsika-run" / "corsika"

    if run_time is None:
        command = [str(corsika_command)]
    else:
        command = [*run_time, str(corsika_command)]

    # Below I do not use the standard context manager because
    # it makes mocking in the tests significantly more difficult
    process = subprocess.Popen(  # pylint: disable=consider-using-with
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        text=True,
    )

    # Capture output until it waits for input
    while True:
        line = process.stdout.readline()
        if not line:
            break
        # Extract the version from the line "NUMBER OF VERSION :  7.7550"
        if "NUMBER OF VERSION" in line:
            version = line.split(":")[1].strip()
            break
        # Check for a specific prompt or indication that the program is waiting for input
        if "DATA CARDS FOR RUN STEERING ARE EXPECTED FROM STANDARD INPUT" in line:
            break

    process.terminate()
    # Check it's a valid version string
    if version and re.match(r"\d+\.\d+", version):
        return version
    try:
        build_opts = get_build_options(run_time)
    except (FileNotFoundError, TypeError, ValueError):
        _logger.warning("Could not get CORSIKA version.")
        return None
    _logger.debug("Getting the CORSIKA version from the build options.")
    return build_opts.get("corsika_version")


def get_build_options(run_time=None):
    """
    Return CORSIKA / sim_telarray build options.

    Expects a build_opts.yml file in the sim_telarray directory.

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    dict
        Build options from build_opts.yml file.
    """
    sim_telarray_path = os.getenv("SIMTOOLS_SIMTEL_PATH")
    if sim_telarray_path is None:
        raise ValueError("SIMTOOLS_SIMTEL_PATH not defined.")

    build_opts_path = Path(sim_telarray_path) / "build_opts.yml"

    if run_time is None:
        try:
            return ascii_handler.collect_data_from_file(build_opts_path)
        except FileNotFoundError as exc:
            raise FileNotFoundError("No build_opts.yml file found.") from exc

    command = [*run_time, "cat", str(build_opts_path)]
    _logger.debug(f"Reading build_opts.yml with command: {command}")

    result = subprocess.run(command, capture_output=True, text=True, check=False)
    if result.returncode:
        raise FileNotFoundError(f"No build_opts.yml file found in container: {result.stderr}")

    try:
        return yaml.safe_load(result.stdout)
    except yaml.YAMLError as exc:
        raise ValueError(f"Error parsing build_opts.yml from container: {exc}") from exc



"""Inspect logs and output for errors and warnings."""

import logging
import re

_logger = logging.getLogger(__name__)


ERROR_PATTERNS = [
    re.compile(r"error", re.IGNORECASE),
    re.compile(r"exception", re.IGNORECASE),
    re.compile(r"traceback", re.IGNORECASE),
    re.compile(r"\b(failed to|has failed)\b", re.IGNORECASE),
    re.compile(r"runtime\s*warning", re.IGNORECASE),
    re.compile(r"segmentation fault", re.IGNORECASE),
]

IGNORE_PATTERNS = [re.compile(r"Falling back to 'utf-8' with errors='ignore'", re.IGNORECASE)]


def inspect(log_text):
    """
    Inspect log text for errors and run-time warnings.

    Ignore any lines containing "INFO::" (to avoid false positives
    like "INFO:: Job error stream ").

    Parameters
    ----------
    log_text: str or list of str
        Text of the log to inspect.

    Returns
    -------
    bool
        True if no errors or warnings are found, False otherwise.
    """
    log_text = log_text if isinstance(log_text, list) else [log_text]

    issues = [
        (lineno, line)
        for txt in log_text
        for lineno, line in enumerate(txt.splitlines(), 1)
        if "INFO::" not in line
        and any(p.search(line) for p in ERROR_PATTERNS)
        and not any(p.search(line) for p in IGNORE_PATTERNS)
    ]

    for lineno, line in issues:
        _logger.error(f"Error or warning found in log at line {lineno}: {line.strip()}")

    return not issues



#!/usr/bin/python3
"""
Plot tabular data read from file or from model parameter database.

Uses a configuration file to define the data to be plotted and all
plotting details.

Command line arguments
----------------------
config_file (str, required)
    Configuration file name for plotting.
output_file (str, required)
    Output file name (without suffix).

Example
-------

Plot tabular data using a configuration file.

.. code-block:: console

    simtools-plot-tabular-data --plot_config config_file_name --output_file output_file_name

"""

import simtools.utils.general as gen
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.constants import PLOT_CONFIG_SCHEMA
from simtools.data_model import schema
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.io import ascii_handler
from simtools.visualization import plot_tables


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Plots tabular data.",
        usage="""simtools-plot-tabular-data --plot_config config_file_name "
                 --output_file output_file_name""",
    )

    config.parser.add_argument(
        "--plot_config",
        help="Plotting configuration file name.",
        type=str,
        required=True,
        default=None,
    )
    config.parser.add_argument(
        "--table_data_path",
        help="Path to the data files (optional). Expect all files to be in the same directory.",
        type=str,
        default=None,
    )
    config.parser.add_argument(
        "--output_file",
        help="Output file name (without suffix)",
        type=str,
        required=True,
    )
    return config.initialize(db_config=True, simulation_model=["telescope"])


def main():
    """Plot tabular data."""
    app_context = startup_application(_parse)

    plot_config = gen.convert_keys_in_dict_to_lowercase(
        schema.validate_dict_using_schema(
            ascii_handler.collect_data_from_file(app_context.args["plot_config"]),
            PLOT_CONFIG_SCHEMA,
        )
    )

    plot_tables.plot(
        config=plot_config["plot"],
        output_file=app_context.io_handler.get_output_file(app_context.args["output_file"]),
        db_config=app_context.db_config,
        data_path=app_context.args.get("table_data_path"),
    )

    MetadataCollector.dump(
        app_context.args,
        app_context.io_handler.get_output_file(app_context.args["output_file"]),
        add_activity_name=True,
    )


if __name__ == "__main__":
    main()



"""IO operations on astropy tables."""

import logging
from pathlib import Path

import astropy.units as u
import h5py
import numpy as np
from astropy.io import fits
from astropy.table import Table, vstack

_logger = logging.getLogger(__name__)


def read_table_list(input_file, table_names, include_indexed_tables=False):
    """
    Read available tables found in the input file.

    If table_counter is True, search for tables with the same name
    but with different suffixes (e.g., "_0", "_1", etc.).

    """
    file_type = read_table_file_type(input_file)
    if file_type == "HDF5":
        return _read_table_list_hdf5(input_file, table_names, include_indexed_tables)
    if file_type == "FITS":
        return _read_table_list_fits(input_file, table_names, include_indexed_tables)
    return None


def _read_table_list_hdf5(input_file, table_names, include_indexed_tables):
    """Read available tables from HDF5 file."""
    datasets = {name: [] for name in table_names}

    def is_indexed_variant(name, base):
        if not name.startswith(f"{base}_"):
            return False
        suffix = name[len(base) + 1 :]
        return suffix.isdigit()

    def visitor(name, obj):
        if not isinstance(obj, h5py.Dataset):
            return

        for base in datasets:
            if name == base or (include_indexed_tables and is_indexed_variant(name, base)):
                datasets[base].append(name)

    with h5py.File(input_file, "r") as f:
        f.visititems(visitor)

    return datasets


def _read_table_list_fits(input_file, table_names, include_indexed_tables):
    """Read available tables from FITS file."""
    datasets = {name: [] for name in table_names}

    with fits.open(input_file) as hdul:
        for hdu in hdul[1:]:
            if not isinstance(hdu, fits.BinTableHDU):
                continue
            name = hdu.name
            if name in table_names:
                datasets[name].append(name)
                continue
            if not include_indexed_tables or "_" not in name:
                continue
            base, _, suffix = name.rpartition("_")
            if base in table_names and suffix.isdigit():
                datasets[base].append(name)

    return datasets


def merge_tables(input_files, input_table_names, output_file):
    """
    Merge multiple astropy tables from different files into a single file.

    Handles multiple tables per file and supports both HDF5 and FITS formats.
    Updates 'file_id' column if present to maintain file origin tracking.

    Parameters
    ----------
    input_files : list of str
        List of input file paths to be merged.
    input_table_names : list of str
        List of table names to be merged from each input file.
    output_file : str
        Path to the output file where the merged data will be saved.

    Returns
    -------
    None
    """
    _logger.info(f"Merging {len(input_files)} files into {output_file}")

    file_type = read_table_file_type(input_files)
    merged_tables = _merge(input_files, input_table_names, file_type, output_file)
    if file_type != "HDF5":
        write_tables(merged_tables, output_file, file_type)


def read_table_file_type(input_files):
    """
    Determine the file type of the input files.

    All input files must be of the same type (either all HDF5 or all FITS).

    Parameters
    ----------
    input_files : list of str
        List of input file paths.

    Returns
    -------
    str
        File type ('HDF5' or 'FITS').
    """
    if not input_files:
        raise ValueError("No input files provided.")
    input_files = [input_files] if isinstance(input_files, str | Path) else input_files

    def get_type(f):
        if f.lower().endswith((".hdf5", ".h5")):
            return "HDF5"
        if f.lower().endswith((".fits", ".fits.gz")):
            return "FITS"
        raise ValueError(f"Unsupported file type: {f}")

    file_types = {get_type(str(f)) for f in input_files}
    if len(file_types) != 1:
        raise ValueError("All input files must be of the same type (either all HDF5 or all FITS)")
    return file_types.pop()


def _merge(input_files, table_names, file_type, output_file, add_file_id_to_table_name=True):
    """
    Merge tables from multiple input files into single tables.

    Parameters
    ----------
    input_files : list of str
        List of input file paths to be merged.
    table_names : list of str
        List of table names to be merged from each input file.
    file_type : str
        Type of the input files ('HDF5' or 'FITS').
    add_file_id_to_table_name : bool, optional
        If True, appends the file index to the table name.

    Returns
    -------
    dict
        Dictionary with table names as keys and merged astropy tables as values.
    """
    merged = {name: [] for name in table_names}
    is_hdf5 = file_type == "HDF5"

    def update_file_id(table, idx):
        if "file_id" in table.colnames:
            table["file_id"] = idx

    def process_table(table, key, idx):
        table_name = f"{key}_{idx}" if add_file_id_to_table_name else key
        update_file_id(table, idx)
        if is_hdf5:
            write_table_in_hdf5(table, output_file, table_name)
            if idx == 0:
                copy_metadata_to_hdf5(input_files[0], output_file, table_name)
        else:
            merged[key].append(table)

    for idx, file in enumerate(input_files):
        tables = read_tables(file, table_names, file_type)
        for key, table in tables.items():
            process_table(table, key, idx)

    if file_type != "HDF5":
        merged = {k: vstack(v, metadata_conflicts="silent") for k, v in merged.items()}

    return merged


def read_tables(file, table_names, file_type=None):
    """
    Read tables from a file.

    Parameters
    ----------
    file : str
        Path to the input file.
    table_names : list of str
        List of table names to read.
    file_type : str
        Type of the input file ('HDF5' or 'FITS').

    Returns
    -------
    dict
        Dictionary with table names as keys and astropy tables as values.
    """
    file_type = file_type or read_table_file_type([file])
    if file_type == "HDF5":
        return {name: read_table_from_hdf5(file, name) for name in table_names}
    if file_type == "FITS":
        return {name: Table.read(file, hdu=name) for name in table_names}
    raise ValueError(f"Unsupported file format: {file_type}. Supported formats are HDF5 and FITS.")


def read_table_from_hdf5(file, table_name):
    """
    Read a single astropy table from an HDF5 file.

    Parameters
    ----------
    file : str or Path
        Path to the input HDF5 file.
    table_name : str
        Name of the table to read.

    Returns
    -------
    astropy.table.Table
        The requested astropy table.
    """
    table = Table.read(file, path=table_name)
    with h5py.File(file, "r") as f:
        dset = f[table_name]
        for col in table.colnames:
            unit_key = f"{col}_unit"
            if unit_key in dset.attrs:
                table[col].unit = u.Unit(dset.attrs[unit_key])
    return table


def write_tables(tables, output_file, overwrite_existing=True, file_type=None):
    """
    Write tables to file (overwriting if exists).

    Parameters
    ----------
    tables : list or dict
        List or Dictionary with astropy tables as values.
    output_file : str or Path
        Path to the output file.
    overwrite_existing : bool
        If True, overwrite the output file if it exists.
    file_type : str
        Type of the output file ('HDF5' or 'FITS').

    Returns
    -------
    None
    """
    output_file = Path(output_file)
    file_type = file_type or read_table_file_type([output_file])
    if output_file.exists():
        if overwrite_existing:
            output_file.unlink()
        else:
            raise FileExistsError(f"Output file {output_file} already exists.")
    hdus = [fits.PrimaryHDU()]
    if isinstance(tables, dict):
        tables = list(tables.values())
    for table in tables:
        _table_name = table.meta.get("EXTNAME")
        _logger.info(f"Writing table {_table_name} of length {len(table)} to {output_file}")
        if file_type == "HDF5":
            write_table_in_hdf5(table, output_file, _table_name)
        if file_type == "FITS":
            hdu = fits.table_to_hdu(table)
            hdu.name = _table_name
            hdus.append(hdu)

    if file_type == "FITS":
        fits.HDUList(hdus).writeto(output_file, checksum=False)


def write_table_in_hdf5(table, output_file, table_name):
    """
    Write or append a single astropy table to an HDF5 file.

    Parameters
    ----------
    table : astropy.table.Table
        The astropy table to write.
    output_file : str or Path
        Path to the output HDF5 file.
    table_name : str
        Name of the table in the HDF5 file.

    Returns
    -------
    None
    """
    for col in table.colnames:
        if table[col].dtype.kind == "U":  # hd5 does not support unicode
            table[col] = table[col].astype("S")

    with h5py.File(output_file, "a") as f:
        data = np.array(table)
        if table_name not in f:
            maxshape = (None, *data.shape[1:])
            dset = f.create_dataset(
                table_name,
                data=data,
                maxshape=maxshape,
                chunks=True,
                compression="gzip",
                compression_opts=4,
            )
            for key, val in table.meta.items():
                dset.attrs[key] = val
            for col in table.colnames:
                unit = getattr(table[col], "unit", None)
                if unit is not None:
                    dset.attrs[f"{col}_unit"] = str(unit)
        else:
            dset = f[table_name]
            dset.resize(dset.shape[0] + data.shape[0], axis=0)
            dset[-data.shape[0] :] = data


def copy_metadata_to_hdf5(src_file, dst_file, table_name):
    """
    Copy metadata (table column meta) from one HDF5 file to another.

    For merging tables, this function ensures that the metadata is preserved.

    Parameters
    ----------
    src_file : str or Path
        Path to the source HDF5 file.
    dst_file : str or Path
        Path to the destination HDF5 file.
    table_name : str
        Name of the table whose metadata is to be copied.
    """
    with h5py.File(src_file, "r") as src, h5py.File(dst_file, "a") as dst:
        meta_name = f"{table_name}.__table_column_meta__"
        if meta_name in src:
            if meta_name in dst:
                del dst[meta_name]  # overwrite if exists
            src.copy(meta_name, dst, name=meta_name)



#!/usr/bin/python3
r"""
    Validate a file or files in a directory using a schema.

    Input files can be metadata, schema, or data files in yaml, json, or ecsv format.

    Command line arguments
    ----------------------
    file_name (str)
      input file to be validated
    model_parameters_directory (str)
        directory with json files of model parameters to be validated
    schema (str)
      schema file (jsonschema format) used for validation
    data_type (str)
        type of input data (allowed types: metadata, schema, data)

    Raises
    ------
    FileNotFoundError
      if file to be validated is not found

    Example
    -------

    .. code-block:: console

        simtools-validate-file-using-schema \\
         --file_name tests/resources/MLTdata-preproduction.meta.yml \\
         --schema simtools/schemas/metadata.metaschema.yml \\
         --data_type metadata

"""

import re
from pathlib import Path

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.constants import MODEL_PARAMETER_SCHEMA_PATH
from simtools.data_model import metadata_collector, schema, validate_data
from simtools.io import ascii_handler


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Validate a file (metadata, schema, or data file) using a schema.",
    )
    config.parser.add_argument(
        "--file_name",
        help="File to be validated (full path or name pattern, e.g., '*.json')",
        default="*.json",
    )
    config.parser.add_argument(
        "--file_directory",
        help=(
            "Directory with files to be validated. "
            "If no schema file is provided, the assumption is that model "
            "parameters are validated and the schema files are taken from "
            f"{MODEL_PARAMETER_SCHEMA_PATH}."
        ),
    )
    config.parser.add_argument("--schema", help="Json schema file", required=False)
    config.parser.add_argument(
        "--data_type",
        help="Type of input data",
        choices=["metadata", "schema", "data", "model_parameter"],
        default="data",
    )
    config.parser.add_argument(
        "--require_exact_data_type",
        help="Require exact data type for validation",
        action="store_true",
    )
    config.parser.add_argument(
        "--ignore_software_version",
        help="Ignore software version check.",
        action="store_true",
    )
    return config.initialize(paths=False)


def _get_schema_file_name(args_dict, data_dict=None):
    """
    Get schema file name from metadata, data dict, or from command line argument.

    Parameters
    ----------
    args_dict (dict)
        command line arguments
    data_dict (dict)
        dictionary with metaschema information

    Returns
    -------
    schema_file: str
        schema file name

    """
    schema_file = args_dict.get("schema")
    if schema_file is None and data_dict is not None:
        schema_file = data_dict.get("meta_schema_url")
    if schema_file is None:
        metadata = metadata_collector.MetadataCollector(
            None, metadata_file_name=args_dict["file_name"]
        )
        schema_file = metadata.get_data_model_schema_file_name()
    return schema_file


def _get_file_list(file_directory=None, file_name="*.json"):
    """Return list of files in a directory."""
    file_list = []
    if file_directory is not None:
        file_list = list(Path(file_directory).rglob(file_name))
        if not file_list:
            raise FileNotFoundError(f"No files found in {file_directory}")
    elif file_name is not None:
        file_list = [file_name]

    return file_list


def validate_dict_using_schema(args_dict, logger):
    """
    Validate a schema file (or several files) given in yaml or json format.

    This function validate all documents in a multi-document YAML file.
    Schema is either given as command line argument, read from the meta_schema_url or from
    the metadata section of the data dictionary.

    """
    for file_name in _get_file_list(args_dict.get("file_directory"), args_dict.get("file_name")):
        try:
            data = ascii_handler.collect_data_from_file(file_name=file_name)
        except FileNotFoundError as exc:
            raise FileNotFoundError(f"Error reading schema file from {file_name}") from exc
        data = data if isinstance(data, list) else [data]
        try:
            for data_dict in data:
                schema.validate_dict_using_schema(
                    data_dict,
                    _get_schema_file_name(args_dict, data_dict),
                    ignore_software_version=args_dict.get("ignore_software_version", False),
                )
        except Exception as exc:
            raise ValueError(f"Validation of file {file_name} failed") from exc
        logger.info(f"Successful validation of file {file_name}")


def validate_data_files(args_dict, logger):
    """Validate data files."""
    if args_dict.get("file_directory") is not None:
        tmp_args_dict = {}
        for file_name in _get_file_list(args_dict.get("file_directory")):
            tmp_args_dict["file_name"] = file_name
            parameter_name = re.sub(r"-\d+\.\d+\.\d+", "", file_name.stem)
            schema_file = schema.get_model_parameter_schema_file(f"{parameter_name}")
            tmp_args_dict["schema"] = schema_file
            tmp_args_dict["data_type"] = "model_parameter"
            tmp_args_dict["require_exact_data_type"] = args_dict["require_exact_data_type"]
            validate_data_file(tmp_args_dict, logger)
    else:
        validate_data_file(args_dict, logger)


def validate_data_file(args_dict, logger):
    """Validate a data file (e.g., in ecsv, json, yaml format)."""
    data_validator = validate_data.DataValidator(
        schema_file=_get_schema_file_name(args_dict),
        data_file=args_dict["file_name"],
        check_exact_data_type=args_dict["require_exact_data_type"],
    )
    data_validator.validate_and_transform(
        is_model_parameter=(args_dict["data_type"].lower() == "model_parameter")
    )

    logger.info(f"Successful validation of data file {args_dict['file_name']}")


def validate_metadata(args_dict, logger):
    """Validate metadata."""
    # metadata_collector runs the metadata validation by default, no need to do anything else
    metadata_collector.MetadataCollector(None, metadata_file_name=args_dict["file_name"])
    logger.info(f"Successful validation of metadata {args_dict['file_name']}")


def main():
    """Validate a file or files in a directory using a schema."""
    app_context = startup_application(_parse)

    if app_context.args["data_type"].lower() == "metadata":
        validate_metadata(app_context.args, app_context.logger)
    elif app_context.args["data_type"].lower() == "schema":
        validate_dict_using_schema(app_context.args, app_context.logger)
    else:
        validate_data_files(app_context.args, app_context.logger)


if __name__ == "__main__":
    main()



"""
Derives the required statistics for a requested set of production parameters through interpolation.

This module provides the `ProductionStatisticsHandler` class, which manages the workflow for
derivation of required number of events for a simulation production using pre-defined metrics.

The module includes functionality to:
- Initialize evaluators for statistical uncertainty calculations based on input parameters.
- Perform interpolation using the initialized evaluators to estimate production statistics at a
query point.
- Write the results of the interpolation to an output file.
"""

import itertools
import logging
from pathlib import Path

import astropy.units as u

from simtools.io import ascii_handler
from simtools.production_configuration.calculate_statistical_uncertainties_grid_point import (
    StatisticalUncertaintyEvaluator,
)
from simtools.production_configuration.interpolation_handler import InterpolationHandler


class ProductionStatisticsHandler:
    """
    Handles the workflow for deriving production statistics.

    This class manages the evaluation of statistical uncertainties from DL2 MC event files
    and performs interpolation to estimate the required number of events for a simulation
    production at a specified query point.
    """

    def __init__(self, args_dict, output_path):
        """
        Initialize the manager with the provided arguments.

        Parameters
        ----------
        args_dict : dict
            Dictionary of command-line arguments.
        output_path : Path
            Path to the directory where the event statistics output file will be saved.
        """
        self.args = args_dict
        self.logger = logging.getLogger(__name__)
        self.output_path = output_path
        self.metrics = ascii_handler.collect_data_from_file(self.args["metrics_file"])
        self.evaluator_instances = []
        self.interpolation_handler = None
        self.grid_points_production = self._load_grid_points_production()

    def _load_grid_points_production(self):
        """Load grid points from the JSON file."""
        grid_points_production_file = self.args["grid_points_production_file"]
        return ascii_handler.collect_data_from_file(grid_points_production_file)

    def initialize_evaluators(self):
        """Initialize StatisticalUncertaintyEvaluator instances for the given grid point."""
        if not (
            self.args["base_path"]
            and self.args["zeniths"]
            and self.args["azimuths"]
            and self.args["nsb"]
            and self.args["offsets"]
        ):
            self.logger.warning("No files read")
            self.logger.warning(f"Base Path: {self.args['base_path']}")
            self.logger.warning(f"Zeniths: {self.args['zeniths']}")
            self.logger.warning(f"Camera offsets: {self.args['offsets']}")
            return

        for zenith, azimuth, nsb, offset in itertools.product(
            self.args["zeniths"], self.args["azimuths"], self.args["nsb"], self.args["offsets"]
        ):
            file_name = self.args["file_name_template"].format(
                zenith=int(zenith),
                azimuth=azimuth,
                nsb=nsb,
                offset=offset,
            )
            file_path = Path(self.args["base_path"]).joinpath(file_name)

            if not file_path.exists():
                self.logger.warning(f"File not found: {file_path}. Skipping.")
                continue

            evaluator = StatisticalUncertaintyEvaluator(
                file_path,
                metrics=self.metrics,
                grid_point=(None, azimuth, zenith, nsb, offset * u.deg),
            )
            evaluator.calculate_metrics()
            self.evaluator_instances.append(evaluator)

    def perform_interpolation(self):
        """Perform interpolation for the query point."""
        if not self.evaluator_instances:
            self.logger.error("No evaluators initialized. Cannot perform interpolation.")
            return None

        self.interpolation_handler = InterpolationHandler(
            self.evaluator_instances,
            metrics=self.metrics,
            grid_points_production=self.grid_points_production,
        )
        qrid_points_with_statistics = []

        interpolated_production_statistics = self.interpolation_handler.interpolate()
        for grid_point, statistics in zip(
            self.grid_points_production, interpolated_production_statistics
        ):
            qrid_points_with_statistics.append(
                {
                    "grid_point": grid_point,
                    "interpolated_production_statistics": float(statistics),
                }
            )
        return qrid_points_with_statistics

    def write_output(self, production_statistics):
        """Write the derived event statistics to a file."""
        output_data = (production_statistics,)
        output_filename = self.args["output_file"]
        self.output_path.mkdir(parents=True, exist_ok=True)
        output_file_path = self.output_path.joinpath(output_filename)
        ascii_handler.write_data_to_file(
            data=output_data,
            output_file=output_file_path,
        )
        self.logger.info(f"Output saved to {self.output_path}")

    def plot_production_statistics_comparison(self):
        """Plot the derived event statistics."""
        ax = self.interpolation_handler.plot_comparison()
        plot_path = self.output_path.joinpath("production_statistics_comparison.png")
        plot_path.parent.mkdir(parents=True, exist_ok=True)
        ax.figure.savefig(plot_path)
        self.logger.info(f"Plot saved to {plot_path}")

    def run(self):
        """Run the scaling and interpolation workflow."""
        self.logger.info(f"Grid Points File: {self.args['grid_points_production_file']}")
        self.logger.info(f"Metrics File: {self.args['metrics_file']}")

        self.initialize_evaluators()
        production_statistics = self.perform_interpolation()
        if self.args.get("plot_production_statistics"):
            self.plot_production_statistics_comparison()

        self.write_output(production_statistics)



#!/usr/bin/python3

r"""
Derive the required number of events for MC productions for a grid of observational conditions.

This application evaluates statistical uncertainties from the analysis of events after the
application of loose gamma/hadron separation cuts, then interpolates the derived number of required
events for the specified grid points provided in a file. The resulting grid points will have the
derived number of required events added.

The metric for the required uncertainty is pre-defined and must be configured via the metrics file.

Command line arguments
----------------------
grid_points_production_file (str, required)
    Path to the file containing grid points. Each grid point should include azimuth, zenith, NSB,
    offset.
metrics_file (str, optional)
    Path to the metrics definition file. Default: 'production_simulation_config_metrics.yml'.
base_path (str, required)
    Path to the directory containing the event files for interpolation (after loose gamma/hadron
    cuts).
file_name_template (str, optional)
    Template for the event file name. Default:
    'prod6_LaPalma-{zenith}deg_gamma_cone.N.Am-4LSTs09MSTs_ID0_reduced.fits'.
plot_production_statistics (flag, optional)
    If provided, plots the production statistics. Default: False.

Example
-------
To evaluate statistical uncertainties and perform interpolation, run the command line script:

.. code-block:: console

    simtools-production-derive-statistics \\
        --grid_points_production_file path/to/grid_points_production.json \\
        --metrics_file "path/to/metrics.yaml" \\
        --base_path path/to/production_event_files/ \\
        --file_name_template "prod6_LaPalma-{zenith}deg\\
            _gamma_cone.N.Am-4LSTs09MSTs_ID0_reduced.fits" \\
        --zeniths 20 40 52 60 \\
        --offsets 0 \\
        --azimuths 180 \\
        --nsb 0.0 \\
        --plot_production_statistics

Output
------
The output will be a file containing the grid points with the derived number of required events
added.
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.production_configuration.derive_production_statistics_handler import (
    ProductionStatisticsHandler,
)


def _parse():
    """Parse command line arguments."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Evaluate statistical uncertainties from DL2 MC event files and interpolate results."
        ),
    )

    config.parser.add_argument(
        "--grid_points_production_file",
        type=str,
        required=True,
        help="Path to the JSON file containing grid points for a production.",
    )
    config.parser.add_argument(
        "--metrics_file",
        required=True,
        type=str,
        default=None,
        help="Metrics definition file. (default: production_simulation_config_metrics.yml)",
    )
    config.parser.add_argument(
        "--base_path",
        type=str,
        required=True,
        help="Path to the DL2 MC event files for interpolation.",
    )
    config.parser.add_argument(
        "--file_name_template",
        required=False,
        type=str,
        default=("prod6_LaPalma-{zenith}deg_gamma_cone.N.Am-4LSTs09MSTs_ID0_reduced.fits"),
        help=("Template for the DL2 MC event file name."),
    )
    config.parser.add_argument(
        "--zeniths",
        required=True,
        nargs="+",
        type=float,
        help="List of zenith angles in deg that describe the supplied DL2 files.",
    )
    config.parser.add_argument(
        "--azimuths",
        required=True,
        nargs="+",
        type=float,
        help="List of azimuth angles in deg that describe the supplied DL2 files.",
    )
    config.parser.add_argument(
        "--nsb",
        required=True,
        nargs="+",
        type=float,
        help="List of nsb values that describe the supplied DL2 files.",
    )
    config.parser.add_argument(
        "--offsets",
        required=True,
        nargs="+",
        type=float,
        help="List of camera offsets in deg that describe the supplied DL2 files.",
    )
    config.parser.add_argument(
        "--plot_production_statistics",
        required=False,
        action="store_true",
        default=False,
        help="Plot production statistics.",
    )

    return config.initialize(db_config=False, output=True)


def main():
    """Run the ProductionStatisticsHandler."""
    app_context = startup_application(_parse)

    manager = ProductionStatisticsHandler(
        app_context.args, output_path=app_context.io_handler.get_output_directory()
    )
    manager.run()


if __name__ == "__main__":
    main()



"""Compare application output to reference output."""

import logging
from pathlib import Path

import numpy as np
from astropy.table import Table

import simtools.utils.general as gen
from simtools.db import db_handler
from simtools.io import ascii_handler
from simtools.testing import assertions

_logger = logging.getLogger(__name__)


def validate_application_output(
    config, from_command_line=None, from_config_file=None, db_config=None
):
    """
    Validate application output against expected output.

    Expected output is defined in configuration file.
    Some tests run only if the model version from the command line
    equals the model version from the configuration file.

    Parameters
    ----------
    config: dict
        dictionary with the configuration for the application test.
    from_command_line: str
        Model version from the command line.
    from_config_file: str
        Model version from the configuration file.

    """
    if "integration_tests" not in config:
        return

    for integration_test in config["integration_tests"]:
        _logger.info(f"Testing application output: {integration_test}")

        if from_command_line == from_config_file:
            _validate_output_files(config, integration_test, db_config)

            if "file_type" in integration_test:
                assert assertions.assert_file_type(
                    integration_test["file_type"],
                    Path(config["configuration"]["output_path"]).joinpath(
                        config["configuration"]["output_file"]
                    ),
                )
        _test_simtel_cfg_files(config, integration_test, from_command_line, from_config_file)


def _validate_output_files(config, integration_test, db_config):
    """Validate output files."""
    if "reference_output_file" in integration_test:
        _validate_reference_output_file(config, integration_test)
    if "test_output_files" in integration_test:
        _validate_output_path_and_file(config, integration_test["test_output_files"])
    if "output_file" in integration_test:
        _validate_output_path_and_file(
            config,
            [{"path_descriptor": "output_path", "file": integration_test["output_file"]}],
        )
    if "model_parameter_validation" in integration_test:
        _validate_model_parameter_json_file(
            config,
            integration_test["model_parameter_validation"],
            db_config,
        )


def _test_simtel_cfg_files(config, integration_test, from_command_line, from_config_file):
    """Test simtel cfg files."""
    cfg_files = integration_test.get("test_simtel_cfg_files", {})
    if isinstance(from_command_line, list):
        sources = from_command_line
    elif isinstance(from_config_file, list):
        sources = from_config_file
    else:
        sources = [from_command_line or from_config_file]
    for version in sources:
        cfg = cfg_files.get(version)
        if cfg:
            _validate_simtel_cfg_files(config, cfg)
            break


def _validate_reference_output_file(config, integration_test):
    """Compare with reference output file."""
    assert compare_files(
        integration_test["reference_output_file"],
        Path(config["configuration"]["output_path"]).joinpath(
            config["configuration"]["output_file"]
        ),
        integration_test.get("tolerance", 1.0e-5),
        integration_test.get("test_columns", None),
    )


def _validate_output_path_and_file(config, integration_file_tests):
    """Check if output paths and files exist."""
    for file_test in integration_file_tests:
        try:
            output_path = config["configuration"][file_test["path_descriptor"]]
        except KeyError as exc:
            raise KeyError(
                f"Path {file_test['path_descriptor']} not found in integration test configuration."
            ) from exc

        output_file_path = Path(output_path) / file_test["file"]
        _logger.info(f"Checking path: {output_file_path}")
        try:
            assert output_file_path.exists()
        except AssertionError as exc:
            raise AssertionError(f"Output file {output_file_path} does not exist. ") from exc

        assert assertions.check_output_from_sim_telarray(output_file_path, file_test)


def _validate_model_parameter_json_file(config, model_parameter_validation, db_config):
    """
    Validate model parameter json file and compare it with a reference parameter from the database.

    Requires database connection to pull the model parameter for a given telescope or site model.

    Parameters
    ----------
    config: dict
        Configuration dictionary.
    model_parameter_validation: dict
        Dictionary with model parameter validation configuration.

    """
    _logger.info(f"Checking model parameter json file: {model_parameter_validation}")
    db = db_handler.DatabaseHandler(db_config=db_config)

    reference_parameter_name = model_parameter_validation.get("reference_parameter_name")

    reference_model_parameter = db.get_model_parameter(
        parameter=reference_parameter_name,
        site=config["configuration"].get("site"),
        array_element_name=config["configuration"].get("telescope"),
        model_version=config["configuration"].get("model_version"),
    )
    parameter_file = (
        Path(config["configuration"]["output_path"])
        / config["configuration"].get("telescope")
        / model_parameter_validation["parameter_file"]
    )
    model_parameter = ascii_handler.collect_data_from_file(parameter_file)
    assert _compare_value_from_parameter_dict(
        model_parameter["value"],
        reference_model_parameter[reference_parameter_name]["value"],
        model_parameter_validation["tolerance"],
    )


def compare_files(file1, file2, tolerance=1.0e-5, test_columns=None):
    """
    Compare two files of file type ecsv, json or yaml.

    Parameters
    ----------
    file1: str
        First file to compare
    file2: str
        Second file to compare
    tolerance: float
        Tolerance for comparing numerical values.
    test_columns: list
        List of columns to compare. If None, all columns are compared.

    Returns
    -------
    bool
        True if the files are equal.

    """
    _file1_suffix = Path(file1).suffix
    _file2_suffix = Path(file2).suffix
    _logger.info("Comparing files: %s and %s", file1, file2)
    if _file1_suffix != _file2_suffix:
        raise ValueError(f"File suffixes do not match: {file1} and {file2}")
    if _file1_suffix == ".ecsv":
        return compare_ecsv_files(file1, file2, tolerance, test_columns)
    if _file1_suffix in (".json", ".yaml", ".yml"):
        return compare_json_or_yaml_files(file1, file2, tolerance)

    _logger.warning(f"Unknown file type for files: {file1} and {file2}")
    return False


def compare_json_or_yaml_files(file1, file2, tolerance=1.0e-2):
    """
    Compare two json or yaml files.

    Take into account float comparison for sim_telarray string-embedded floats.
    Allow differences in 'schema_version' field.

    Parameters
    ----------
    file1: str
        First file to compare
    file2: str
        Second file to compare
    tolerance: float
        Tolerance for comparing numerical values.

    Returns
    -------
    bool
        True if the files are equal.

    """
    data1 = ascii_handler.collect_data_from_file(file1)
    data2 = ascii_handler.collect_data_from_file(file2)
    data1.pop("schema_version", None)
    data2.pop("schema_version", None)

    _logger.debug(f"Comparing json/yaml files: {file1} and {file2}")

    if data1 == data2:
        return True

    if data1.keys() != data2.keys():
        _logger.error(f"Keys do not match: {data1.keys()} and {data2.keys()}")
        return False
    _comparison = all(
        (
            _compare_value_from_parameter_dict(data1[k], data2[k], tolerance)
            if k == "value"
            else data1[k] == data2[k]
        )
        for k in data1
    )
    if not _comparison:
        _logger.error(f"Values do not match: {data1} and {data2} (tolerance: {tolerance})")
    return _comparison


def _compare_value_from_parameter_dict(data1, data2, tolerance=1.0e-5):
    """Compare value fields given in different formats."""

    def _as_list(value):
        if isinstance(value, str):
            return gen.convert_string_to_list(value)
        if isinstance(value, list | np.ndarray):
            return value
        return [value]

    _logger.info(f"Comparing values: {data1} and {data2} (tolerance: {tolerance})")

    _as_list_1 = _as_list(data1)
    _as_list_2 = _as_list(data2)
    if isinstance(_as_list_1, str):
        return _as_list_1 == _as_list_2
    return np.allclose(_as_list_1, _as_list_2, rtol=tolerance)


def compare_ecsv_files(file1, file2, tolerance=1.0e-5, test_columns=None):
    """
    Compare two ecsv files.

    The comparison is successful if:

    - same number of rows
    - numerical values in columns are close

    The comparison can be restricted to a subset of columns with some additional
    cuts applied. This is configured through the test_columns parameter. This is
    a list of dictionaries, where each dictionary contains the following
    key-value pairs:
    - test_column_name: column name to compare.
    - cut_column_name: column for filtering.
    - cut_condition: condition for filtering.

    Parameters
    ----------
    file1: str
        First file to compare
    file2: str
        Second file to compare
    tolerance: float
        Tolerance for comparing numerical values.
    test_columns: list
        List of columns to compare. If None, all columns are compared.

    """
    _logger.info(f"Comparing files: {file1} and {file2}")
    table1 = Table.read(file1, format="ascii.ecsv")
    table2 = Table.read(file2, format="ascii.ecsv")

    if test_columns is None:
        test_columns = [{"test_column_name": col} for col in table1.colnames]

    def generate_mask(table, column, condition):
        """Generate a boolean mask based on the condition (note the usage of eval)."""
        return (
            eval(f"table['{column}'] {condition}")  # pylint: disable=eval-used
            if condition
            else np.ones(len(table), dtype=bool)
        )

    for col_dict in test_columns:
        col_name = col_dict["test_column_name"]
        mask1 = generate_mask(
            table1, col_dict.get("cut_column_name", ""), col_dict.get("cut_condition", "")
        )
        mask2 = generate_mask(
            table2, col_dict.get("cut_column_name", ""), col_dict.get("cut_condition", "")
        )
        table1_masked, table2_masked = table1[mask1], table2[mask2]

        if len(table1_masked) != len(table2_masked):
            return False

        if np.issubdtype(table1_masked[col_name].dtype, np.floating):
            if not np.allclose(table1_masked[col_name], table2_masked[col_name], rtol=tolerance):
                _logger.warning(f"Column {col_name} outside of relative tolerance {tolerance}")
                return False

    return True


def _validate_simtel_cfg_files(config, simtel_cfg_file):
    """
    Check sim_telarray configuration files and compare with reference file.

    Note the finetuned naming of configuration files by simtools.

    """
    reference_file = Path(simtel_cfg_file)
    test_file = (
        Path(config["configuration"]["output_path"])
        / f"model/{config['configuration']['model_version']}"
        / reference_file.name.replace("_test", f"_{config['configuration']['label']}")
    )
    _logger.info(
        f"Comparing simtel cfg files: {reference_file} and {test_file} "
        f"for model version {config['configuration']['model_version']}"
    )
    assert _compare_simtel_cfg_files(reference_file, test_file)


def _compare_simtel_cfg_files(reference_file, test_file):
    """
    Compare two sim_telarray configuration files.

    Line-by-line string comparison. Requires similar sequence of
    parameters in the files. Ignore lines containing 'config_release'
    (as it contains the simtools package version).

    Parameters
    ----------
    reference_file: Path
        Reference sim_telarray configuration file.
    test_file: Path
        Test sim_telarray configuration file.

    Returns
    -------
    bool
        True if the files are equal.

    """
    with open(reference_file, encoding="utf-8") as f1, open(test_file, encoding="utf-8") as f2:
        reference_cfg = [line.rstrip() for line in f1 if line.strip()]
        test_cfg = [line.rstrip() for line in f2 if line.strip()]

    if len(reference_cfg) != len(test_cfg):
        _logger.error(
            f"Line counts differ: {reference_file} ({len(reference_cfg)} lines), "
            f"{test_file} ({len(test_cfg)} lines)."
        )
        return False

    for ref_line, test_line in zip(reference_cfg, test_cfg):
        if any(ignore in ref_line for ignore in ("config_release", "Label", "simtools_version")):
            continue
        if ref_line != test_line:
            _logger.error(
                f"Configuration files {reference_file} and {test_file} do not match: "
                f"'{ref_line}' and '{test_line}'"
            )
            return False

    return True



r"""
Derive cosmic-ray trigger rates for a single telescope or an array of telescopes.

Uses simulated background events (e.g. from proton primaries) to calculate the trigger rates.
Input is reduced event data generated from simulations for the given configuration.


Command line arguments
----------------------
event_data_file (str, required)
    Event data file containing reduced event data.
array_layout_name (list, optional)
    Name of the array layout to use for the simulation.
telescope_ids (str, optional)
    Path to a file containing telescope configurations.
plot_histograms (bool, optional)
    Plot histograms of the event data.
model_version (str, optional)
    Version of the simulation model to use.
site (str, optional)
    Name of the site where the simulation is being run.


Example
-------

Derive trigger rates for the South Alpha layout:

.. code-block:: console

    simtools-derive-trigger-rates \\
        --site South \\
        --model_version 6.0.0 \\
        --event_data_file /path/to/event_data_file.h5 \\
        --array_layout_name alpha\\
        --plot_histograms

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.telescope_trigger_rates import telescope_trigger_rates


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Derive trigger rates for a single telescope or an array of telescopes.",
    )
    config.parser.add_argument(
        "--event_data_file",
        type=str,
        required=True,
        help="Event data file containing reduced event data.",
    )
    config.parser.add_argument(
        "--telescope_ids",
        type=str,
        required=False,
        help="Path to a file containing telescope configurations.",
    )
    config.parser.add_argument(
        "--plot_histograms",
        help="Plot histograms of the event data.",
        action="store_true",
        default=False,
    )
    return config.initialize(
        db_config=True,
        output=True,
        simulation_model=[
            "site",
            "model_version",
            "layout",
        ],
    )


def main():
    """Derive trigger rates for a single telescope or an array of telescopes."""
    app_context = startup_application(_parse)

    telescope_trigger_rates(app_context.args, app_context.db_config)


if __name__ == "__main__":
    main()



"""Trigger rate calculation for telescopes and arrays of telescopes."""

import logging

import numpy as np
from astropy import units as u
from ctao_cr_spectra.definitions import IRFDOC_PROTON_SPECTRUM

from simtools.io import ascii_handler, io_handler
from simtools.layout.array_layout_utils import get_array_elements_from_db_for_layouts
from simtools.simtel.simtel_io_event_histograms import SimtelIOEventHistograms
from simtools.visualization import plot_simtel_event_histograms

_logger = logging.getLogger(__name__)


def telescope_trigger_rates(args_dict, db_config):
    """
    Calculate trigger rates for single telescopes or arrays of telescopes.

    Main function to read event data, fill histograms, and derive trigger rates.


    """
    if args_dict.get("array_layout_name"):
        telescope_configs = get_array_elements_from_db_for_layouts(
            args_dict["array_layout_name"],
            args_dict.get("site"),
            args_dict.get("model_version"),
            db_config,
        )
    else:
        telescope_configs = ascii_handler.collect_data_from_file(args_dict["telescope_ids"])[
            "telescope_configs"
        ]

    for array_name, telescope_ids in telescope_configs.items():
        _logger.info(
            f"Processing file: {args_dict['event_data_file']} with telescope config: {array_name}"
        )
        histograms = SimtelIOEventHistograms(
            args_dict["event_data_file"], array_name=array_name, telescope_list=telescope_ids
        )
        histograms.fill()

        _calculate_trigger_rates(histograms, array_name)

        if args_dict["plot_histograms"]:
            plot_simtel_event_histograms.plot(
                histograms.histograms,
                output_path=io_handler.IOHandler().get_output_directory(),
                array_name=array_name,
            )


def _calculate_trigger_rates(histograms, array_name):
    """
    Calculate trigger rates from the filled histograms.

    Missing

    - custom definition of energy spectra

    """
    efficiency = histograms.histograms.get("energy_eff", {}).get("histogram")
    energy_axis = histograms.histograms.get("energy_eff", {}).get("bin_edges")

    cr_spectrum = get_cosmic_ray_spectrum()
    _logger.info(f"Cosmic ray spectrum: {cr_spectrum}")
    e_min = energy_axis[:-1] * u.TeV
    e_max = energy_axis[1:] * u.TeV
    cr_rates = (
        np.array(
            [
                cr_spectrum.integrate_energy(e1, e2).decompose(bases=[u.s, u.cm, u.sr]).value
                for e1, e2 in zip(e_min, e_max)
            ]
        )
        * histograms.file_info["scatter_area"].to("cm2").value
        * histograms.file_info["solid_angle"].to("sr").value
        * u.Hz
    )
    trigger_rates = efficiency * cr_rates
    trigger_rate = np.sum(trigger_rates, axis=0)

    _logger.info(f"Scatter area from MC: {histograms.file_info['scatter_area'].to('m2')}")
    _logger.info(f"Solid angle from MC: {histograms.file_info['solid_angle']}")
    _logger.info(f"Trigger rate for {array_name} array: {trigger_rate.to('Hz')}")

    histograms.histograms["cr_rates_mc"] = histograms.get_histogram_definition(
        histogram=cr_rates.value,
        bin_edges=energy_axis,
        title="Cosmic Ray Rates (MC)",
        axis_titles=["Energy (TeV)", "Cosmic Ray Rate (Hz)"],
        plot_scales={"x": "log", "y": "log"},
    )
    histograms.histograms["trigger_rates"] = histograms.get_histogram_definition(
        histogram=trigger_rates.value,
        bin_edges=energy_axis,
        title="Trigger Rates (MC)",
        axis_titles=["Energy (TeV)", "Trigger Rate (Hz)"],
        plot_scales={"x": "log", "y": "log"},
    )

    return cr_rates, trigger_rates, trigger_rate


def get_cosmic_ray_spectrum():
    """
    Return the cosmic ray spectrum.

    To be extended in future to read a larger variety of spectra.

    Returns
    -------
    astropy.units.Quantity
        Cosmic ray spectrum.
    """
    return IRFDOC_PROTON_SPECTRUM



r"""
Generate a new simulation model production and update tables and model parameters.

This script is used to maintain the simulation model repository. It allows to create
new production tables by copying an existing base version and applies modifications
to production tables and model parameters as provided in a configuration file (see
the 'info.yml' examples in the simulation models repository).

Two main use cases are covered by this script:

1. full update: create a complete new set of production tables (e.g. for new major or minor
   versions of the simulation models). This will copy all production tables from the source
   directory and apply the modifications to the tables that are listed in the modifications file.
   If the full update is based on a previous patch update, the full history of changes is applied
   iteratively until the last base version is reached.

2. patch update: create a set of new production tables including the changes defined in the
   modifications file. No unmodified tables are copied. For new production tables with patch
   modifications, the key-value pair 'base_model_version: <base_model version>' is added.

Both use cases will also apply the modifications to the model parameters as defined in the
modifications file.

Example
-------

The following example applies a patch update with changes defined in a YAML file.

.. code-block:: console

    simtools-maintain-simulation-model-add-new-production \\
        --model_path ../simulation-models-dev/simulation-models/ \\
        --model_version 6.0.2

"""

from pathlib import Path

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model import model_repository


def _parse():
    """Parse command line arguments."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Generate a new simulation model production",
    )
    return config.initialize(db_config=False, output=False, simulation_model=["model_version"])


def main():
    """Generate a new simulation model production."""
    app_context = startup_application(_parse)

    model_repository.generate_new_production(
        model_version=app_context.args["model_version"],
        simulation_models_path=Path(app_context.args["model_path"]),
    )


if __name__ == "__main__":
    main()



"""Validation of data using schema."""

import logging
import os
import re
from pathlib import Path

import numpy as np
from astropy import units as u
from astropy.table import Column, Table, unique
from astropy.utils.diff import report_diff_values

import simtools.utils.general as gen
from simtools.data_model import schema
from simtools.io import ascii_handler
from simtools.utils import names, value_conversion


class DataValidator:
    """
    Validate data for type and units following a describing schema; converts or transform data.

    Data can be of table or dict format (internally, all data is converted to astropy tables).

    Parameters
    ----------
    schema_file: Path
        Schema file describing input data and transformations.
    data_file: Path
        Input data file.
    data_table: astropy.table
        Input data table.
    data_dict: dict
        Input data dict.
    check_exact_data_type: bool
        Check for exact data type (default: True).

    """

    def __init__(
        self,
        schema_file=None,
        data_file=None,
        data_table=None,
        data_dict=None,
        check_exact_data_type=True,
    ):
        """Initialize validation class and read required reference data columns."""
        self._logger = logging.getLogger(__name__)

        self.data_file_name = data_file
        self.schema_file_name = schema_file
        self._data_description = None
        self.data_dict = data_dict
        self.data_table = data_table
        self.check_exact_data_type = check_exact_data_type

    def validate_and_transform(self, is_model_parameter=False, lists_as_strings=False):
        """
        Validate data and data file.

        Parameters
        ----------
        is_model_parameter: bool
            This is a model parameter (add some data preparation)
        lists_as_strings: bool
            Convert lists to strings (as needed for model parameters)

        Returns
        -------
        data: dict or astropy.table
            Data dict or table

        Raises
        ------
        TypeError
            if no data or data table is available

        """
        if self.data_file_name:
            self.validate_data_file(is_model_parameter)
        if isinstance(self.data_dict, dict):
            return self._validate_data_dict(is_model_parameter, lists_as_strings)
        if isinstance(self.data_table, Table):
            return self._validate_data_table()
        self._logger.error("No data or data table to validate")
        raise TypeError

    def validate_data_file(self, is_model_parameter=None):
        """
        Open data file and read data from file.

        Doing this successfully is understood as file validation.

        Parameters
        ----------
        is_model_parameter: bool
            This is a model parameter file.
        """
        try:
            if Path(self.data_file_name).suffix in (".yml", ".yaml", ".json"):
                self.data_dict = ascii_handler.collect_data_from_file(self.data_file_name)
                self._logger.info(f"Validating data from: {self.data_file_name}")
            else:
                self.data_table = Table.read(self.data_file_name, guess=True, delimiter=r"\s")
                self._logger.info(f"Validating tabled data from: {self.data_file_name}")
        except (AttributeError, TypeError):
            pass
        if is_model_parameter:
            self.validate_parameter_and_file_name()

    def validate_parameter_and_file_name(self):
        """
        Validate model parameter file name.

        Expect that the following convention is used:

        - file name starts with the parameter name
        - file name ends with parameter version string

        """
        file_stem = Path(self.data_file_name).stem
        param = self.data_dict.get("parameter")
        param_version = self.data_dict.get("parameter_version")
        if not file_stem.startswith(param):
            raise ValueError(f"Mismatch: parameter '{param}' vs. file '{file_stem}'.")

        if param_version and not file_stem.endswith(param_version):
            raise ValueError(f"Mismatch: version '{param_version}' vs. file '{file_stem}'.")

        if param_version is None:
            self._logger.warning(f"File '{file_stem}' has no parameter version defined.")

    @staticmethod
    def validate_model_parameter(par_dict):
        """
        Validate a simulation model parameter (static method).

        Parameters
        ----------
        par_dict: dict
            Data dictionary

        Returns
        -------
        dict
            Validated data dictionary
        """
        data_validator = DataValidator(
            schema_file=schema.get_model_parameter_schema_file(f"{par_dict['parameter']}"),
            data_dict=par_dict,
            check_exact_data_type=False,
        )
        return data_validator.validate_and_transform(is_model_parameter=True)

    def _validate_data_dict(self, is_model_parameter=False, lists_as_strings=False):
        """
        Validate values in a dictionary.

        Handles different types of naming in data dicts (using 'name' or 'parameter'
        keys for name fields).

        Parameters
        ----------
        is_model_parameter: bool
            This is a model parameter (add some data preparation)
        lists_as_strings: bool
            Convert lists to strings (as needed for model parameters)

        Raises
        ------
        KeyError
            if data dict does not contain a 'name' or 'parameter' key.

        """
        if is_model_parameter:
            self._prepare_model_parameter()

        self._data_description = self._read_validation_schema(
            self.schema_file_name,
            # use 0.1.0 as default; this corresponds to the first definition of the schema
            self.data_dict.get("model_parameter_schema_version", "0.1.0"),
        )

        value_as_list, unit_as_list = self._get_value_and_units_as_lists()

        for index, (value, unit) in enumerate(zip(value_as_list, unit_as_list)):
            try:
                value_as_list[index], unit_as_list[index] = self._validate_value_and_unit(
                    value, unit, index
                )
            except TypeError as ex:
                raise TypeError(
                    f"Error validating dictionary using {self.schema_file_name}"
                ) from ex

        if len(value_as_list) == 1:
            self.data_dict["value"], self.data_dict["unit"] = value_as_list[0], unit_as_list[0]
        else:
            self.data_dict["value"], self.data_dict["unit"] = value_as_list, unit_as_list

        if self.data_dict.get("instrument"):
            if self.data_dict["instrument"] == self.data_dict["site"]:  # site parameters
                names.validate_site_name(self.data_dict["site"])
            else:
                names.validate_array_element_name(self.data_dict["instrument"])
                self._check_site_and_array_element_consistency(
                    self.data_dict.get("instrument"), self.data_dict.get("site")
                )

        for version_string in ("version", "parameter_version", "model_version"):
            self._check_version_string(self.data_dict.get(version_string))

        if lists_as_strings:
            self._convert_results_to_model_format()

        return self.data_dict

    def _validate_value_and_unit(self, value, unit, index):
        """
        Validate value, unit, and perform type checking and conversions.

        Take into account different data types and allow to use json_schema for testing.
        """
        if self._get_data_description(index).get("type", None) == "dict":
            schema.validate_dict_using_schema(
                data=self.data_dict["value"],
                json_schema=self._get_data_description(index).get("json_schema"),
            )
        else:
            self._check_data_type(np.array(value).dtype, index, value)

        if self._get_data_description(index).get("type", None) not in ("string", "dict", "file"):
            self._check_for_not_a_number(value, index)
            value, unit = self._check_and_convert_units(value, unit, index)
            for range_type in ("allowed_range", "required_range"):
                self._check_range(index, np.nanmin(value), np.nanmax(value), range_type)
        return value, unit

    def _get_value_and_units_as_lists(self):
        """
        Convert value and unit to lists if required.

        Ignore unit field in data_dict if value is a astropy.Quantity.
        Note the complications from astropy.Units, where a single value is of np.ndarray type.

        Returns
        -------
        list
            value as list
        list
            unit as list
        """
        target_unit = self.data_dict["unit"]
        value, unit = value_conversion.split_value_and_unit(self.data_dict["value"])

        if not isinstance(value, list | np.ndarray):
            value, unit = [value], [unit]
        if not isinstance(target_unit, list | np.ndarray):
            target_unit = [target_unit] * len(value)

        target_unit = [None if unit == "null" else unit for unit in target_unit]
        conversion_factor = [
            1 if v is None else u.Unit(v).to(u.Unit(t)) for v, t in zip(unit, target_unit)
        ]
        try:
            return [
                v * c if not isinstance(v, bool) and not isinstance(v, dict) else v
                for v, c in zip(value, conversion_factor)
            ], target_unit
        except TypeError:
            return [None], target_unit

    def _validate_data_table(self):
        """Validate tabulated data."""
        try:
            self._data_description = self._read_validation_schema(self.schema_file_name)[0].get(
                "table_columns", None
            )
        except IndexError:
            self._logger.error(f"Error reading validation schema from {self.schema_file_name}")
            raise

        if self._data_description is not None:
            self._validate_data_columns()
            self._check_data_for_duplicates()
            self._sort_data()
        return self.data_table

    def _validate_data_columns(self):
        """
        Validate that data columns.

        This includes:

        - required data columns are available
        - columns are in the correct units (if necessary apply a unit conversion)
        - ranges (minimum, maximum) are correct.

        This is not applied to columns of type 'string'.

        """
        self._check_required_columns()

        for col_name in self.data_table.colnames:
            col = self.data_table[col_name]
            if not self._get_data_description(col_name, status_test=True):
                continue
            if not np.issubdtype(col.dtype, np.number):
                continue
            self._check_for_not_a_number(col.data, col_name)
            self._check_data_type(col.dtype, col_name)
            self.data_table[col_name] = col.to(u.Unit(self._get_reference_unit(col_name)))
            self._check_range(col_name, np.nanmin(col.data), np.nanmax(col.data), "allowed_range")
            self._check_range(col_name, np.nanmin(col.data), np.nanmax(col.data), "required_range")

    def _check_required_columns(self):
        """
        Check that all required data columns are available in the input data table.

        Raises
        ------
        KeyError
            if a required data column is missing

        """
        for entry in self._data_description:
            if entry.get("required", False):
                if entry["name"] in self.data_table.columns:
                    self._logger.debug(f"Found required data column {entry['name']}")
                else:
                    raise KeyError(f"Missing required column {entry['name']}")

    def _sort_data(self):
        """
        Sort data according to one data column (if required by any column attribute).

        Data is either sorted or reverse sorted.

        Raises
        ------
        AttributeError
            if no table is defined for sorting

        """
        _columns_by_which_to_sort = []
        _columns_by_which_to_reverse_sort = []
        for entry in self._data_description:
            if "input_processing" in entry:
                if "sort" in entry["input_processing"]:
                    _columns_by_which_to_sort.append(entry["name"])
                elif "reversesort" in entry["input_processing"]:
                    _columns_by_which_to_reverse_sort.append(entry["name"])

        if len(_columns_by_which_to_sort) > 0:
            self._logger.debug(f"Sorting data columns: {_columns_by_which_to_sort}")
            try:
                self.data_table.sort(_columns_by_which_to_sort)
            except AttributeError:
                self._logger.error("No data table defined for sorting")
                raise
        elif len(_columns_by_which_to_reverse_sort) > 0:
            self._logger.debug(f"Reverse sorting data columns: {_columns_by_which_to_reverse_sort}")
            try:
                self.data_table.sort(_columns_by_which_to_reverse_sort, reverse=True)
            except AttributeError:
                self._logger.error("No data table defined for reverse sorting")
                raise

    def _check_data_for_duplicates(self):
        """
        Remove duplicates from data columns as defined in the data columns description.

        Raises
        ------
            if row values are different for those rows with duplications in the data columns to be
            checked for unique values.

        """
        _column_with_unique_requirement = self._get_unique_column_requirement()
        if len(_column_with_unique_requirement) == 0:
            self._logger.debug("No data columns with unique value requirement")
            return
        _data_table_unique_for_key_column = unique(
            self.data_table, keys=_column_with_unique_requirement
        )
        _data_table_unique_for_all_columns = unique(self.data_table, keys=None)
        with open(os.devnull, "w", encoding="utf-8") as devnull:
            if report_diff_values(
                _data_table_unique_for_key_column,
                _data_table_unique_for_all_columns,
                fileobj=devnull,
            ):
                self.data_table = unique(self.data_table)
            else:
                raise ValueError(
                    "Failed removal of duplication for column "
                    f"{_column_with_unique_requirement}, values are not unique"
                )

    def _get_unique_column_requirement(self):
        """
        Return data column name with unique value requirement.

        Returns
        -------
        list
            list of data column with unique value requirement

        """
        _unique_required_column = []

        for entry in self._data_description:
            if "input_processing" in entry and "remove_duplicates" in entry["input_processing"]:
                self._logger.debug(f"Removing duplicates for column {entry['name']}")
                _unique_required_column.append(entry["name"])

        self._logger.debug(f"Unique required columns: {_unique_required_column}")
        return _unique_required_column

    def _get_reference_unit(self, column_name):
        """
        Return reference column unit. Includes correct treatment of dimensionless units.

        Parameters
        ----------
        column_name: str
            column name of reference data column

        Returns
        -------
        astro.unit
            unit for reference column

        Raises
        ------
        KeyError
            if column name is not found in reference data columns

        """
        reference_unit = self._get_data_description(column_name).get("unit", None)
        if reference_unit in ("dimensionless", None, ""):
            return u.dimensionless_unscaled

        return u.Unit(reference_unit)

    def _check_data_type(self, dtype, column_name, value=None):
        """
        Check column data type.

        Parameters
        ----------
        dtype: numpy.dtype
            data type
        column_name: str
            column name
        value: value
            value to be tested (optional)

        Raises
        ------
        TypeError
            if data type is not correct

        """
        reference_dtype = self._get_data_description(column_name).get("type", None)
        if not gen.validate_data_type(
            reference_dtype=reference_dtype,
            value=value,
            dtype=dtype,
            allow_subtypes=(not self.check_exact_data_type),
        ):
            self._logger.error(
                f"Invalid data type in column '{column_name}'. "
                f"Expected type '{reference_dtype}', found '{dtype}' "
                f"(exact type: {self.check_exact_data_type})"
            )
            raise TypeError

    def _check_for_not_a_number(self, data, col_name):
        """
        Check that column values are finite and not NaN.

        Parameters
        ----------
        data: value or numpy.ndarray
            data to be tested
        col_name: str
            column name

        Returns
        -------
        bool
            if at least one column value is NaN or Inf.

        Raises
        ------
        ValueError
            if at least one column value is NaN or Inf.

        """
        if isinstance(data, str):
            return True
        if not isinstance(data, np.ndarray):
            data = np.array(data)

        if np.isnan(data).any():
            self._logger.info(f"Column {col_name} contains NaN.")
        if np.isinf(data).any():
            self._logger.info(f"Column {col_name} contains infinite value.")

        entry = self._get_data_description(col_name)
        if "allow_nan" in entry.get("input_processing", {}):
            return np.isnan(data).any() or np.isinf(data).any()

        if np.isnan(data).any() or np.isinf(data).any():
            raise ValueError("NaN or Inf values found in data")

        return False

    @staticmethod
    def _is_dimensionless(unit):
        """
        Check if unit is dimensionless, None, or empty.

        Parameters
        ----------
        unit: str
            unit of data column

        Returns
        -------
        bool
            True if unit is dimensionless, None, or empty
        """
        return unit in ("dimensionless", None, "")

    def _check_and_convert_units(self, data, unit, col_name):
        """
        Check that input data have an allowed unit.

        Convert to reference unit (e.g., Angstrom to nm).

        Note on dimensionless columns:

        - should be given in unit descriptor as unit: ''
        - be forgiving and assume that in cases no unit is given in the data files
          means that it should be dimensionless (e.g., for a efficiency)

        Parameters
        ----------
        data: astropy.column, Quantity, list, value
            data to be converted
        unit: str
            unit of data column (read from column or Quantity if possible)
        col_name: str
            column name

        Returns
        -------
        data: astropy.column, Quantity, list, value
            unit-converted data

        Raises
        ------
        u.core.UnitConversionError
            If unit conversions fails

        """
        self._rate_limited_logger(col_name, f"Checking data column '{col_name}'")

        reference_unit = self._get_reference_unit(col_name)
        try:
            column_unit = data.unit
        except AttributeError:
            column_unit = unit

        if self._is_dimensionless(column_unit) and self._is_dimensionless(reference_unit):
            return data, u.dimensionless_unscaled

        self._rate_limited_logger(
            col_name,
            f"Data column '{col_name}' with reference unit "
            f"'{reference_unit}' and data unit '{column_unit}'",
        )
        try:
            if isinstance(data, u.Quantity | Column):
                return data.to(reference_unit), reference_unit

            if isinstance(data, list | np.ndarray):
                return self._check_and_convert_units_for_list(data, column_unit, reference_unit)

            # ensure that the data type is preserved (e.g., integers)
            return (type(data)(u.Unit(column_unit).to(reference_unit) * data), reference_unit)
        except (u.core.UnitConversionError, ValueError) as exc:
            self._logger.error(
                f"Invalid unit in data column '{col_name}'. "
                f"Expected type '{reference_unit}', found '{column_unit}'"
            )
            raise u.core.UnitConversionError from exc

    def _check_and_convert_units_for_list(self, data, column_unit, reference_unit):
        """
        Check and convert units data in a list or or numpy array.

        Takes into account that data can be dimensionless (with unit 'None', 'dimensionless'
        or '').

        Parameters
        ----------
        data: list
            list of data
        column_unit: str
            unit of data column
        reference_unit: str
            reference unit

        Returns
        -------
        list
            converted data

        """
        return [
            (
                u.Unit(_to_unit).to(reference_unit) * d
                if _to_unit not in (None, "dimensionless", "")
                else d
            )
            for d, _to_unit in zip(data, column_unit)
        ], reference_unit

    def _check_range(self, col_name, col_min, col_max, range_type="allowed_range"):
        """
        Check that column data is within allowed range or required range.

        Assumes that column and ranges have the same units.

        Parameters
        ----------
        col_name: string
            column name
        col_min: float
            minimum value of data column
        col_max: float
            maximum value of data column
        range_type: string
            column range type (either 'allowed_range' or 'required_range')

        Raises
        ------
        ValueError
            if columns are not in the required range
        KeyError
            if requested columns cannot be found or
            if there is now defined of required or allowed
            range columns

        """
        self._rate_limited_logger(
            col_name, f"Checking data in column '{col_name}' for '{range_type}'"
        )

        if range_type not in ("allowed_range", "required_range"):
            raise KeyError("Allowed range types are 'allowed_range', 'required_range'")

        _entry = self._get_data_description(col_name)
        if range_type not in _entry:
            return

        if not self._interval_check(
            (col_min, col_max),
            (_entry[range_type].get("min", -np.inf), _entry[range_type].get("max", np.inf)),
            range_type,
        ):
            raise ValueError(
                f"Value for column '{col_name}' out of range. "
                f"([{col_min}, {col_max}], {range_type}: "
                f"[{_entry[range_type].get('min', -np.inf)}, "
                f"{_entry[range_type].get('max', np.inf)}])"
            )

    def _rate_limited_logger(self, col_name, message, max_logs=10):
        """
        Log debug messages at a limited rate defined by a numerical column name.

        Parameters
        ----------
        col_name: str or int
            Column name or index to limit the rate of logging.
        message: str
            Message to log.
        max_logs: int
            Maximum number of logging messages to be printed.
        """
        try:
            col_index = int(col_name)
            if col_index < max_logs:
                self._logger.debug(message)
        except (ValueError, TypeError):
            self._logger.debug(message)

    @staticmethod
    def _interval_check(data, axis_range, range_type):
        """
        Range checking for a given set of data.

        Check that values are inside allowed range or interval. This(range_type='allowed_range')
        or span at least the given interval (range_type='required_range').

        Parameters
        ----------
        data: tuple
            min and max of data
        axis_range: tuple
            allowed or required min max
        range_type: string
            column range type (either 'allowed_range' or 'required_range')

        Returns
        -------
        boolean
            True if range test is passed

        """
        if range_type == "allowed_range":
            if data[0] >= axis_range[0] and data[1] <= axis_range[1]:
                return True
        if range_type == "required_range":
            if data[0] <= axis_range[0] and data[1] >= axis_range[1]:
                return True

        return False

    def _read_validation_schema(self, schema_file, schema_version=None):
        """
        Read validation schema from file.

        The schema file can be a yaml file with several documents, each document
        describing a different schema version. Returns first document if no
        schema version is requested.

        Parameters
        ----------
        schema_file: Path
            Schema file describing input data.
        schema_version: str
            Version of the schema to be used.

        Returns
        -------
        dict
           validation schema

        Raises
        ------
        KeyError
            if 'data' can not be read from dict in schema file
        ValueError
            if schema version is not found in schema file
        """
        schema_data = ascii_handler.collect_data_from_file(file_name=schema_file)
        entries = schema_data if isinstance(schema_data, list) else [schema_data]

        for entry in entries:
            if not schema_version or entry.get("schema_version") == schema_version:
                try:
                    return entry["data"]
                except KeyError as exc:
                    raise KeyError(f"Error reading validation schema from {schema_file}") from exc
        raise ValueError(
            f"Schema version {schema_version} not found in schema file {schema_file}. "
        )

    def _get_data_description(self, column_name=None, status_test=False):
        """
        Return data description as provided by the schema file.

        For tables (type: 'data_table'), return the description of
        the column named 'column_name'. For other types, return
        all data descriptions.
        For columns named 'colX' return the Xth column in the reference data.

        Parameters
        ----------
        column_name: str
            Column name.
        status_test: bool
            Test if reference column exists.

        Returns
        -------
        dict
            Reference schema column (for status_test==False).
        bool
            True if reference column exists (for status_test==True).

        Raises
        ------
        IndexError
            If data column is not found.

        """
        self._rate_limited_logger(
            column_name,
            f"Getting reference data column {column_name} from schema {self._data_description}",
        )
        try:
            return (
                self._data_description[column_name]
                if not status_test
                else (
                    self._data_description[column_name] is not None
                    and len(self._data_description) > 0
                )
            )
        except IndexError as exc:
            if len(self._data_description) == 1:  # all columns are described by the same schema
                return self._data_description[0]
            self._logger.error(
                f"Data column '{column_name}' not found in reference column definition"
            )
            raise exc
        except TypeError:
            pass  # column_name is not an integer

        _index = 0
        if bool(re.match(r"^col\d$", column_name)):
            _index = int(column_name[3:])
            _entry = self._data_description
        else:
            _entry = [item for item in self._data_description if item["name"] == column_name]
        if status_test:
            return len(_entry) > 0
        try:
            return _entry[_index]
        except IndexError:
            self._logger.error(
                f"Data column '{column_name}' not found in reference column definition"
            )
            raise

    def _prepare_model_parameter(self):
        """
        Apply data preparation for model parameters.

        Converts strings to numerical values or lists of values, if required.

        """
        self.data_dict["value"], _ = value_conversion.split_value_and_unit(
            self.data_dict["value"],
            "int" in self.data_dict.get("type", "float"),
        )
        if isinstance(self.data_dict["unit"], str):
            self.data_dict["unit"] = gen.convert_string_to_list(
                self.data_dict["unit"], force_comma_separation=True
            )

    def _convert_results_to_model_format(self):
        """
        Convert results to model format.

        Convert lists to strings (as needed for model parameters).
        """
        value = self.data_dict["value"]
        if isinstance(value, list):
            self.data_dict["value"] = gen.convert_list_to_string(value)
        if isinstance(self.data_dict["unit"], list):
            self.data_dict["unit"] = gen.convert_list_to_string(self.data_dict["unit"])

    def _check_version_string(self, version):
        """
        Check that version string follows semantic versioning.

        Parameters
        ----------
        version: str
            version string

        Raises
        ------
        ValueError
            if version string does not follow semantic versioning

        """
        if version is None:
            return
        semver_regex = r"^\d+\.\d+\.\d+(-[0-9A-Za-z.-]+)?(\+[0-9A-Za-z.-]+)?$"
        if not re.match(semver_regex, version):
            raise ValueError(f"Invalid version string '{version}'")
        self._logger.debug(f"Valid version string '{version}'")

    def _check_site_and_array_element_consistency(self, instrument, site):
        """
        Check that site and array element names are consistent.

        An example for an inconsistency is 'LSTN' at site 'South'
        """
        if not all([instrument, site]) or "OBS" in instrument:
            return

        def to_sorted_list(value):
            """Return value as sorted list."""
            return [value] if isinstance(value, str) else sorted(value)

        instrument_site = to_sorted_list(names.get_site_from_array_element_name(instrument))
        site = to_sorted_list(site)

        if instrument_site != site:
            raise ValueError(
                f"Site '{site}' and instrument site '{instrument_site}' are inconsistent."
            )



#!/usr/bin/python3
"""Plot tabular data."""

import logging
from pathlib import Path

import numpy as np
from astropy.table import Table

import simtools.utils.general as gen
from simtools.constants import SCHEMA_PATH
from simtools.db import db_handler
from simtools.io import ascii_handler, legacy_data_handler
from simtools.visualization import visualize

_logger = logging.getLogger(__name__)


def plot(config, output_file, db_config=None, data_path=None):
    """
    Plot tabular data from data or from model parameter files.

    Parameters
    ----------
    config: dict
        Configuration dictionary for plotting.
    output_file: str
        Output file.
    db_config: dict, optional
        Database configuration dictionary for accessing the model parameter database.
    data_path: Path or str, optional
        Path to the data files (optional). Expect all files to be in the same directory.
    """
    data = read_table_data(config, db_config, data_path)

    fig = visualize.plot_1d(
        data,
        **config,
    )
    visualize.save_figure(fig, output_file)

    return output_file


def read_table_data(config, db_config, data_path=None):
    """
    Read table data from file or parameter database.

    Parameters
    ----------
    config: dict
        Configuration dictionary for plotting.
    db_config: dict
        Database configuration dictionary for accessing the model parameter database.
    data_path: Path or str, optional
        Path to the data files (optional). Expect all files to be in the same directory.

    Returns
    -------
    Dict
        Dict with table data (astropy tables).
    """
    data = {}

    for _config in config["tables"]:
        if "parameter" in _config:
            table = _read_table_from_model_database(_config, db_config)
        elif "file_name" in _config:
            file_name = (
                _config["file_name"]
                if data_path is None or _config.get("ignore_table_data_path", False)
                else Path(data_path) / _config["file_name"]
            )
            _logger.info(f"Reading tabular data from {file_name}")
            if "legacy" in _config.get("type", ""):
                table = legacy_data_handler.read_legacy_data_as_table(file_name, _config["type"])
            else:
                table = Table.read(file_name, format="ascii.ecsv")
        else:
            raise ValueError("No table data defined in configuration.")

        data[_get_plotting_label(_config, data)] = _process_table_data(table, _config)

    return data


def _get_plotting_label(config, data):
    """Get a label for plotting based on the configuration."""
    label = config.get("label", f"{config.get('column_x')} vs {config.get('column_y')}")
    if label in data:
        index = 1
        while f"{label} ({index})" in data:
            index += 1
        label = f"{label} ({index})"
    return label


def _process_table_data(table, _config):
    """Process table data based on configuration."""
    if _config.get("normalize_y"):
        table[_config["column_y"]] = table[_config["column_y"]] / table[_config["column_y"]].max()
    if _config.get("select_values"):
        table = _select_values_from_table(
            table,
            _config["select_values"]["column_name"],
            _config["select_values"]["value"],
        )

    return gen.get_structure_array_from_table(
        table,
        [
            _config["column_x"],
            _config["column_y"],
            _config.get("column_x_err"),
            _config.get("column_y_err"),
        ],
    )


def _read_table_from_model_database(table_config, db_config):
    """
    Read table data from model parameter database.

    Parameters
    ----------
    table_config: dict
        Configuration dictionary for table data.

    Returns
    -------
    Table
        Astropy table
    """
    db = db_handler.DatabaseHandler(db_config=db_config)
    return db.export_model_file(
        parameter=table_config["parameter"],
        site=table_config["site"],
        array_element_name=table_config.get("telescope"),
        parameter_version=table_config.get("parameter_version"),
        model_version=table_config.get("model_version"),
        export_file_as_table=True,
    )


def _select_values_from_table(table, column_name, value):
    """Return a table with only the rows where column_name == value."""
    return table[np.isclose(table[column_name], value)]


def _filter_config_by_plot_type(config, plot_type):
    """Filter a configuration based on plot type."""
    if plot_type != "all" and config.get("type") != plot_type:
        return False
    return True


def _validate_config_columns(config, valid_columns, logger):
    """Validate that all required columns in a config exist and have valid data."""
    for table_config in config.get("tables", []):
        required_cols = [table_config.get("column_x"), table_config.get("column_y")]
        if not all(col in valid_columns for col in required_cols if col):
            missing_cols = [col for col in required_cols if col not in valid_columns]
            logger.info(
                f"Skipping plot config {config.get('type')}: "
                f"Missing valid data in columns: {missing_cols}"
            )
            return False
    return True


def _get_valid_columns(table):
    """Return columns that exist and have valid data (not all NaN)."""
    valid_columns = []
    for col in table.colnames:
        if not all(np.isnan(table[col])):
            valid_columns.append(col)
    return valid_columns


def generate_plot_configurations(
    parameter, parameter_version, site, telescope, output_path, plot_type, db_config
):
    """
    Generate plot configurations for a model parameter from schema files.

    Parameters
    ----------
    parameter: str
        Model parameter name.
    parameter_version: str
        Parameter version.
    site: str
        Site name.
    telescope: str
        Telescope name.
    output_path: str or Path
        Output path for the plots.
    plot_type: str
        Plot type or "all" for all plots.
    db_config: dict
        Database configuration.

    Returns
    -------
    tuple
        Tuple containing a list of plot configurations and a list of output file names.
        Return None, if no plot configurations are found.
    """
    logger = logging.getLogger(__name__)

    # Get schema configuration
    schema = gen.change_dict_keys_case(
        ascii_handler.collect_data_from_file(
            file_name=SCHEMA_PATH / "model_parameters" / f"{parameter}.schema.yml"
        )
    )
    configs = schema.get("plot_configuration")
    if not configs:
        return None

    # Get data table and determine valid columns
    table = _read_table_from_model_database(
        {
            "parameter": parameter,
            "site": site,
            "telescope": telescope,
            "parameter_version": parameter_version,
        },
        db_config=db_config,
    )
    valid_columns = _get_valid_columns(table)

    # Filter configs based on plot type and column validity
    valid_configs = []
    for config in configs:
        if not _filter_config_by_plot_type(config, plot_type):
            continue

        if _validate_config_columns(config, valid_columns, logger):
            valid_configs.append(config)

    if not valid_configs:
        if plot_type != "all":
            logger.warning("No valid plot config found.")
        return None

    # Generate output files
    output_files = []
    for _config in valid_configs:
        for _table in _config.get("tables", []):
            _table["parameter_version"] = parameter_version
            _table["site"] = site
            _table["telescope"] = telescope

        output_files.append(
            _generate_output_file_name(
                parameter=parameter,
                parameter_version=parameter_version,
                site=site,
                telescope=telescope,
                plot_type=_config.get("type"),
                output_path=output_path,
            )
        )

    return valid_configs, output_files


def _generate_output_file_name(
    parameter,
    parameter_version,
    site,
    telescope,
    plot_type,
    output_path=None,
    file_extension=".pdf",
):
    """Generate output file name based on table file and appendix."""
    parts = [parameter, parameter_version, site]
    if telescope:
        parts.append(telescope)
    if plot_type != parameter:
        parts.append(plot_type)
    filename = "_".join(parts) + file_extension

    return Path(output_path) / filename if output_path else Path(filename)



#!/usr/bin/python3
"""
    Read array element positions from file and add them to model repository.

    This is an application for experts and should not be used by the general user.
    Reading of input is fine-tuned to the array element files as provided by CTAO.

    Command line arguments

    input : str
        File containing a table of array element positions.
    repository_path : str
        Path of local copy of model parameter repository.
    parameter_version : str
        Parameter version.
    site : str
        Observatory site.
    coordinate_system : str
        Coordinate system of array element positions (ground or utm).

    Examples
    --------
    Add array element positions to repository:

    .. code-block:: console

        simtools-write-array-element-positions-to-repository \
            --input /path/to/positions.txt \
            --repository_path /path/to/repository \
            --parameter_version 0.1.0 \
            --coordinate_system ground \
            --site North

"""

from pathlib import Path

import astropy.table

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.data_model.model_data_writer import ModelDataWriter
from simtools.io import ascii_handler
from simtools.model.array_model import ArrayModel


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Add array element positions to model parameter repository",
    )
    config.parser.add_argument(
        "--input",
        help="File containing a table of array element positions.",
        required=False,
    )
    config.parser.add_argument(
        "--repository_path",
        help="Output path to model parameter repository.",
        type=Path,
        required=False,
    )
    config.parser.add_argument(
        "--coordinate_system",
        help="Coordinate system of array element positions (utm or ground).",
        default="ground",
        required=False,
        type=str,
        choices=["ground", "utm"],
    )

    return config.initialize(db_config=True, simulation_model=["site", "parameter_version"])


def write_utm_array_elements_to_repository(args_dict, logger):
    """
    Write UTM position of array elements to model repository.

    Read array element positions from file. The ecsv row definition might
    include telescope_name or asset_code and sequence_number.

    Parameters
    ----------
    args_dict : dict
        Command line arguments.
    app_context.logger : app_context.logger
        app_context.logger object.

    """
    array_elements = astropy.table.Table.read(args_dict["input"])
    for row in array_elements:
        instrument = (
            row["telescope_name"]
            if "telescope_name" in array_elements.colnames
            else f"{row['asset_code']}-{row['sequence_number']}"
        )
        output_path = Path(args_dict["repository_path"]) / f"{instrument}"
        output_path.mkdir(parents=True, exist_ok=True)
        logger.info(f"Writing array element positions (utm) to {output_path}")

        ModelDataWriter.dump_model_parameter(
            parameter_name="array_element_position_utm",
            instrument=instrument,
            value=f"{row['utm_east']} {row['utm_north']} {row['altitude']}",
            parameter_version=args_dict["parameter_version"],
            output_path=output_path,
            output_file="array_element_position_utm.json",
        )


def write_ground_array_elements_to_repository(args_dict, db_config, logger):
    """
    Write ground position of array elements to model repository.

    Parameters
    ----------
    args_dict : dict
        Command line arguments.
    db_config : dict
        Database configuration.
    logger : logger
        logger object.

    """
    array_model = ArrayModel(
        db_config=db_config,
        model_version=None,
        site=args_dict["site"],
        array_elements=args_dict["input"],
    )
    for element_name, data in array_model.array_elements.items():
        output_path = Path(args_dict["repository_path"]) / f"{element_name}"
        output_path.mkdir(parents=True, exist_ok=True)
        logger.info(f"Writing array element positions (ground) to {output_path}")
        ascii_handler.write_data_to_file(
            data=data,
            output_file=output_path / "array_element_position_ground.json",
            sort_keys=False,
            numpy_types=True,
        )


def main():
    """Application main."""
    app_context = startup_application(_parse)

    if app_context.args["coordinate_system"] == "utm":
        write_utm_array_elements_to_repository(app_context.args, app_context.logger)
    elif app_context.args["coordinate_system"] == "ground":
        write_ground_array_elements_to_repository(
            app_context.args, app_context.db_config, app_context.logger
        )
    else:
        app_context.logger.error("Invalid coordinate system. Allowed are 'utm' and 'ground'.")
        raise ValueError


if __name__ == "__main__":
    main()
